import io
import uuid
from plynx.utils.remote import get_driver


def get_file_stream(file_path, preview=False, file_type=None):
    content = get_driver().get_contents_handler(file_path)
    content_stream = io.BytesIO()
    content.get_contents_to_file(content_stream)
    content_stream.seek(0)
    return content_stream


def upload_file_stream(fp, file_path=None, seek=True):
    if seek:
        fp.seek(0)
    if file_path is None:
        file_path = str(uuid.uuid1())
    content = get_driver().get_contents_handler(file_path)
    content.set_contents_from_file(fp)
    return file_path


def remove(file_path):
    content = get_driver().get_contents_handler(file_path)
    content.remove()



import logging
from plynx.db.node_cache import NodeCache
from plynx.utils.db_connector import get_db_connector
from plynx.constants import NodeRunningStatus


class NodeCacheManager(object):
    """The Node cache interface.

    The cache is defined by Node's
        - original_node_id
        - inputs
        - parameters
    """

    @staticmethod
    def get(node, user_id):
        """Pull NodeCache if exists.

        Args:
            node        (Node):             Node object
            user_id     (ObjectId, str):    User ID

        Return:
            (NodeCache)     NodeCache or None
        """
        key = NodeCache.generate_key(node, user_id)
        db_node_cache = get_db_connector().node_cache.find({
            'key': key,
            'removed': {'$ne': True}
        }).sort('insertion_date', -1).limit(1)
        caches = list(db_node_cache)
        if len(caches):
            return NodeCache.from_dict(caches[0])
        else:
            return None

    # TODO Demo: remove user_id
    @staticmethod
    def post(node, graph_id, user_id):
        """Create NodeCache instance in the database.

        Args:
            node        (Node):             Node object
            graph_id    (ObjectId, str):    Graph ID
            user_id     (ObjectId, str):    User ID

        Return:
            True if cache saved else False
        """
        assert node.node_running_status == NodeRunningStatus.SUCCESS, \
            'Only Nodes with status SUCCESS can be cached'
        node_cache = NodeCache.instantiate(node=node, graph_id=graph_id, user_id=user_id)
        try:
            node_cache.save()
        except Exception as e:
            logging.error('Could not save cache: `{}`'.format(e))
            return False
        return True

    @staticmethod
    def _make_query(start_datetime=None, end_datetime=None, non_protected_only=False):
        """Make sample query.

        Args:
            start_datetime  (datetime, None):   Start datetime or None if selecting from beginning
            end_datetime    (datetime, None):   End datetime or None if selecting until now

        Return:
            Iterator on the list of dict-like objects
        """
        and_query = []

        insertion_query = {}
        if start_datetime:
            insertion_query['$gte'] = start_datetime
        if end_datetime:
            insertion_query['$lt'] = end_datetime
        if insertion_query:
            and_query.append({'insertion_date': insertion_query})

        if non_protected_only:
            and_query.append({'protected': {'$ne': True}})

        return {'$and': and_query} if and_query else {}

    @staticmethod
    def get_list(start_datetime=None, end_datetime=None, non_protected_only=False):
        """List of NodeCache objects.

        Args:
            start_datetime  (datetime, None):   Start datetime or None if selecting from beginning
            end_datetime    (datetime, None):   End datetime or None if selecting until now

        Return:
            Iterator on the list of dict-like objects
        """
        return get_db_connector().node_cache.find(NodeCacheManager._make_query(start_datetime, end_datetime, non_protected_only))

    @staticmethod
    def clean_up():
        """Remove NodeCache objects with flag `removed` set
        """
        return get_db_connector().node_cache.remove({'removed': True})



from pymongo import ReturnDocument
from past.builtins import basestring
from collections import OrderedDict
from plynx.db.node import Node
from plynx.constants import NodeRunningStatus, Collections, NodeStatus
from plynx.utils.common import to_object_id, parse_search_string
from plynx.utils.db_connector import get_db_connector

_PROPERTIES_TO_GET_FROM_SUBS = ['node_running_status', 'logs', 'outputs', 'cache_url']


class NodeCollectionManager(object):
    """NodeCollectionManager contains all the operations to work with Nodes in the database."""

    def __init__(self, collection):
        super(NodeCollectionManager, self).__init__()

        self.collection = collection

    def get_db_objects(
            self,
            status='',
            node_kinds=None,
            search='',
            per_page=20,
            offset=0,
            user_id=None,
            ):
        """Get subset of the Objects.

        Args:
            status              (str, None):                Node Running Status
            search              (str, None):                Search pattern
            per_page            (int):                      Number of Nodes per page
            offset              (int):                      Offset

        Return:
            (list of dict)  List of Nodes in dict format
        """
        if status and isinstance(status, basestring):
            status = [status]
        if node_kinds and isinstance(node_kinds, basestring):
            node_kinds = [node_kinds]

        aggregate_list = []
        search_parameters, search_string = parse_search_string(search)

        # Match
        and_query = {}
        if node_kinds:
            and_query['kind'] = {'$in': node_kinds}
        if status:
            and_query['node_status'] = {'$in': status}
        if search_string:
            and_query['$text'] = {'$search': search_string}
        if 'original_node_id' in search_parameters:
            and_query['original_node_id'] = to_object_id(search_parameters['original_node_id'])
        if len(and_query):
            aggregate_list.append({"$match": and_query})

        # Join with users
        aggregate_list.append({
            '$lookup': {
                'from': 'users',
                'localField': 'author',
                'foreignField': '_id',
                'as': '_user'
            }
        })
        # rm password hash
        aggregate_list.append({
          "$project": {
            "_user.password_hash": 0,
          }
        })

        # Match username
        and_query = {}
        if 'author' in search_parameters:
            and_query['_user.username'] = search_parameters['author']

        if len(and_query):
            aggregate_list.append({"$match": and_query})

        # sort
        sort_dict = OrderedDict()
        if 'sort' in search_parameters:
            # TODO more sort options
            if search_parameters['sort'] == 'starred':
                sort_dict['starred'] = -1
        sort_dict['insertion_date'] = -1

        aggregate_list.append({
            "$sort": sort_dict
            }
        )
        aggregate_list.append({
            "$addFields": {
                '_readonly': {'$ne': ["$author", to_object_id(user_id)]},
            }
        })
        # counts and pagination
        aggregate_list.append({
            '$facet': {
                "metadata": [{"$count": "total"}],
                "list": [{"$skip": int(offset)}, {"$limit": int(per_page)}],
            }
        })

        return next(get_db_connector()[self.collection].aggregate(aggregate_list), None)

    def get_db_objects_by_ids(self, ids, collection=None):
        """Find all the Objects with a given IDs.

        Args:
            ids    (list of ObjectID):  Object Ids
        """
        db_objects = get_db_connector()[collection or self.collection].find({
            '_id': {
                '$in': list(ids)
            }
        })

        return list(db_objects)

    def _update_sub_nodes_fields(self, sub_nodes_dicts, reference_node_id, target_props, reference_collection=None):
        if not sub_nodes_dicts:
            return
        reference_collection = reference_collection or self.collection
        id_to_updated_node_dict = {}
        upd_node_ids = set(map(lambda node_dict: node_dict[reference_node_id], sub_nodes_dicts))
        for upd_node_dict in self.get_db_objects_by_ids(upd_node_ids, collection=reference_collection):
            id_to_updated_node_dict[upd_node_dict['_id']] = upd_node_dict
        for sub_node_dict in sub_nodes_dicts:
            if sub_node_dict[reference_node_id] not in id_to_updated_node_dict:
                continue
            for prop in target_props:
                sub_node_dict[prop] = id_to_updated_node_dict[sub_node_dict[reference_node_id]][prop]

    def get_db_node(self, node_id, user_id=None):
        """Get dict representation of a Node.

        Args:
            node_id     (ObjectId, str):        Object ID
            user_id     (str, ObjectId, None):  User ID

        Return:
            (dict)  dict representation of the Object
        """
        res = self.get_db_object(node_id, user_id)
        if not res:
            return res

        sub_nodes_dicts = None
        for parameter in res['parameters']:
            if parameter['name'] == '_nodes':
                sub_nodes_dicts = parameter['value']['value']
                break

        # TODO join collections using database capabilities
        if self.collection == Collections.RUNS:
            self._update_sub_nodes_fields(sub_nodes_dicts, '_id', _PROPERTIES_TO_GET_FROM_SUBS)
        self._update_sub_nodes_fields(sub_nodes_dicts, 'original_node_id', ['node_status'], reference_collection=Collections.TEMPLATES)

        return res

    def get_db_object(self, object_id, user_id=None):
        """Get dict representation of an Object.

        Args:
            object_id   (ObjectId, str):        Object ID
            user_id     (str, ObjectId, None):  User ID

        Return:
            (dict)  dict representation of the Object
        """
        res = get_db_connector()[self.collection].find_one({'_id': to_object_id(object_id)})
        if not res:
            return res

        res['_readonly'] = (user_id != to_object_id(res['author']))

        return res

    @staticmethod
    def _transplant_node(node, new_node):
        if new_node._id == node.original_node_id:
            return node
        new_node.apply_properties(node)
        new_node.original_node_id = new_node._id
        new_node.parent_node_id = new_node.successor_node_id = None
        new_node._id = node._id
        return new_node

    def upgrade_sub_nodes(self, main_node):
        """Upgrade deprecated Nodes.

        The function does not change the original graph in the database.

        Return:
            (int):  Number of upgraded Nodes
        """
        assert self.collection == Collections.TEMPLATES
        sub_nodes = main_node.get_parameter_by_name('_nodes').value.value
        node_ids = set(
            [node.original_node_id for node in sub_nodes]
        )
        db_nodes = self.get_db_objects_by_ids(node_ids)
        new_node_db_mapping = {}

        for db_node in db_nodes:
            original_node_id = db_node['_id']
            new_db_node = db_node
            if original_node_id not in new_node_db_mapping:
                while new_db_node['node_status'] != NodeStatus.READY and 'successor_node_id' in new_db_node and new_db_node['successor_node_id']:
                    n = self.get_db_node(new_db_node['successor_node_id'])
                    if n:
                        new_db_node = n
                    else:
                        break
                new_node_db_mapping[original_node_id] = new_db_node

        new_nodes = [
            NodeCollectionManager._transplant_node(
                node,
                Node.from_dict(new_node_db_mapping[to_object_id(node.original_node_id)])
            ) for node in sub_nodes]

        upgraded_nodes_count = sum(
            1 for node, new_node in zip(sub_nodes, new_nodes) if node.original_node_id != new_node.original_node_id
        )

        main_node.get_parameter_by_name('_nodes').value.value = new_nodes
        return upgraded_nodes_count

    def pick_node(self, kinds):
        node = get_db_connector()[self.collection].find_one_and_update(
            {
                '$and': [
                    {
                        'kind': {
                            '$in': kinds,
                        }
                    },
                    {
                        'node_running_status': {
                            '$in': [
                                NodeRunningStatus.READY,
                                NodeRunningStatus.IN_QUEUE,
                            ]
                        }
                    },
                ],
            },
            {
                '$set': {
                    'node_running_status': NodeRunningStatus.RUNNING
                }
            },
            return_document=ReturnDocument.AFTER
        )
        return node



from __future__ import print_function
import argparse
from collections import namedtuple
from plynx import __version__
from plynx.utils.config import get_config, set_parameter
from plynx.service.worker import run_worker
from plynx.service.users import run_users
from plynx.service.cache import run_cache
from plynx.service.execute import run_execute
from plynx.utils.logs import set_logging_level


Arg = namedtuple(
    'Arg',
    ['flags', 'help', 'action', 'default', 'nargs', 'type', 'levels', 'required']
)
Arg.__new__.__defaults__ = (None, None, None, None, None, None)

_config = get_config()


def api(args):
    # lazy load because web initializes too many variables
    from plynx.web.common import run_api    # noqa: E402
    set_logging_level(args.pop('verbose'))
    run_api(**args)


def cache(args):
    set_logging_level(args.pop('verbose'))
    run_cache(**args)


def worker(args):
    set_logging_level(args.pop('verbose'))
    run_worker(**args)


def users(args):
    run_users(**args)


def version(args):
    print(__version__)


def execute(args):
    set_logging_level(args.pop('verbose'))
    run_execute(**args)


class CLIFactory(object):
    ARGS = {
        # Shared
        'verbose': Arg(
            ('-v', '--verbose'),
            help='Set logging output more verbose',
            default=0,
            action='count',
            ),
        'mode': Arg(
            ('--mode',),
            help='Mode',
            type=str),

        # Worker
        'kinds': Arg(
            ("-e", "--kinds"),
            help="Kinds the worker is subscribed to",
            default=_config.worker.kinds,
            action='append',
            levels=['worker', 'kinds'],
            ),

        # MongoConfig
        'db_host': Arg(
            ('--db-host',),
            help='Database host',
            default=_config.db.host,
            type=str,
            levels=['mongodb', 'host'],
            ),
        'db_port': Arg(
            ('--db-port',),
            help='Database port',
            default=_config.db.port,
            type=str,
            levels=['mongodb', 'port'],
            ),
        'db_user': Arg(
            ('--db-user',),
            help='Database username',
            default=None,
            type=str,
            levels=['mongodb', 'user'],
            ),
        'db_password': Arg(
            ('--db-password',),
            help='Database password',
            default=None,
            type=str,
            levels=['mongodb', 'password'],
            ),

        # StorageConfig
        'storage_scheme': Arg(
            ('--storage-scheme',),
            help='Storage scheme',
            default=_config.storage.scheme,
            type=str,
            levels=['storage', 'scheme'],
            ),
        'storage_prefix': Arg(
            ('--storage-prefix',),
            help='Storage prefix',
            default=_config.storage.prefix,
            type=str,
            levels=['storage', 'prefix'],
            ),
        'credential_path': Arg(
            ('--credential-path',),
            help='Path to the credentials, i.e. $GOOGLE_APPLICATION_CREDENTIALS',
            default=_config.storage.credential_path,
            type=str,
            levels=['storage', 'credential_path'],
            ),

        # AuthConfig
        'secret_key': Arg(
            ('--secret-key',),
            help='Secret Key path (used in auth). If not given, a single user mode will be used.',
            default=None,
            type=str,
            levels=['auth', 'secret_key'],
            ),

        # WebConfig
        'endpoint': Arg(
            ('--endpoint',),
            help='Endpoint of api',
            default=_config.web.endpoint,
            type=str,
            levels=['web', 'endpoint'],
            ),

        # Users
        'username': Arg(
            ('--username',),
            help='Username of the user, required to create/delete a user',
            type=str),
        'password': Arg(
            ('--password',),
            help='Password of the user, required to create a user '
                 'without --use_random_password',
            type=str),

        # Cache
        'start_datetime': Arg(
            ("--start-datetime", ),
            "End limit time based operation",
            default=None,
            type=str,
            ),
        'end_datetime': Arg(
            ("--end-datetime", ),
            "End limit time based operation",
            default=None,
            type=str,
            ),
        'yes': Arg(
            ("-y", "--yes"),
            "Do not prompt to confirm reset. Use with care!",
            "store_true",
            default=False),

        # Execute
        'filename': Arg(
            ('-f', '--filename'),
            help='Path to file',
            required=True,
            type=str),
    }

    SUBPARSERS = (
        {
            'func': worker,
            'help': 'Run Worker',
            'args': ('verbose', 'db_host', 'db_port', 'db_user', 'db_password', 'kinds',
                     'storage_scheme', 'storage_prefix', 'credential_path'),
        }, {
            'func': api,
            'help': 'Run api server',
            'args': ('verbose', 'secret_key', 'endpoint',
                     'db_host', 'db_port', 'db_user', 'db_password',
                     'storage_scheme', 'storage_prefix', 'credential_path'),
        }, {
            'func': version,
            'help': "Show the version",
            'args': tuple(),
        }, {
            'func': users,
            'help': "Users cli utils",
            'args': ('mode', 'username', 'password', 'db_host'),
        }, {
            'func': cache,
            'help': "Cache cli utils",
            'args': ('verbose', 'mode', 'start_datetime', 'end_datetime', 'yes'),
        }, {
            'func': execute,
            'help': "Execute single node",
            'args': ('verbose', 'filename', 'storage_prefix'),
        },
    )

    @classmethod
    def parse_global_config_parameters(cls, args):
        remove = []
        for k, v in args.items():
            if cls.ARGS[k].levels:
                if v is not None:
                    # Do not set parameter if it is undefined
                    set_parameter(cls.ARGS[k].levels, v)
                remove.append(k)
        for k in remove:
            del args[k]

    @classmethod
    def get_parser(cls):
        parser = argparse.ArgumentParser()
        subparsers = parser.add_subparsers(
            help='sub-command help', dest='subcommand')
        subparsers.required = True

        for sub in cls.SUBPARSERS:
            sp = subparsers.add_parser(sub['func'].__name__, help=sub['help'])
            for arg in sub['args']:
                arg = cls.ARGS[arg]
                kwargs = {
                    f: getattr(arg, f)
                    for f in arg._fields if f not in ['flags', 'levels'] and getattr(arg, f) is not None
                }
                sp.add_argument(*arg.flags, **kwargs)
            sp.set_defaults(func=sub['func'], args=sub['args'])

        return parser


def get_parser():
    return CLIFactory.get_parser()



# flake8: noqa
from plynx.constants.collections import Collections
from plynx.constants.node_enums import NodeRunningStatus, NodeStatus, \
    NodePostAction, NodePostStatus, JobReturnStatus, NodeClonePolicy, NodeVirtualCollection, SpecialNodeId
from plynx.constants.parameter_types import ParameterTypes
from plynx.constants.resource_enums import ResourcePostStatus, NodeResources
from plynx.constants.validation_enums import ValidationTargetType, ValidationCode



import json
from plynx.db.node import Node
from plynx.base import hub
from plynx.utils.common import parse_search_string


def _enhance_list_item(raw_item):
    if raw_item['_type'] == 'Group':
        # TODO proper checking
        items = []
        for raw_subitem in raw_item['items']:
            items.append(_enhance_list_item(raw_subitem))
        raw_item['items'] = items
        return raw_item
    # check if the node is valid
    node = Node.from_dict(raw_item)
    return node.to_dict()


class StaticListHub(hub.BaseHub):
    def __init__(self, filename):
        super(StaticListHub, self).__init__()

        self.list_of_nodes = []

        with open(filename) as f:
            data_list = json.load(f)
            for raw_item in data_list:
                self.list_of_nodes.append(_enhance_list_item(raw_item))

    def search(self, query):
        # TODO use search_parameters
        # TODO should parse_search_string be removed from nodes_collection?
        _, search_string = parse_search_string(query.search)

        def filter_func(raw_node):
            return len(search_string) == 0 or \
                search_string.upper() in raw_node['title'].upper()
        res = list(filter(
            filter_func,
            self.list_of_nodes
        ))
        return {
            'list': res,
            'metadata': [
                {
                    'total': len(res)
                },
            ],
        }



class ResourcePostStatus:
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'


class NodeResources:
    INPUT = 'inputs'
    OUTPUT = 'outputs'
    CLOUD_INPUT = 'cloud_inputs'
    CLOUD_OUTPUT = 'cloud_outputs'
    PARAM = 'params'
    LOG = 'logs'



import os
import shutil
from plynx.utils.remote.base import ContentsHandlerBase, RemoteBase


class ContentsHandlerFile(ContentsHandlerBase):
    def __init__(self, remote, path):
        super(ContentsHandlerFile, self).__init__(remote)
        self.path = path

    def get_contents_to_file(self, file_obj):
        with open(self.path, 'rb') as f:
            shutil.copyfileobj(f, file_obj)

    def set_contents_from_file(self, file_obj):
        with open(self.path, 'wb') as f:
            shutil.copyfileobj(file_obj, f)

    def remove(self):
        os.remove(self.path)

    def exists(self):
        return os.path.exists(self.path)


class RemoteFile(RemoteBase):
    def __init__(self, storage_config):
        super(RemoteFile, self).__init__(ContentsHandlerFile, storage_config)

        # make sure directory exists
        if not os.path.exists(self._storage_config.prefix):
            os.makedirs(self._storage_config.prefix)



from subprocess import Popen
import os
import signal
import logging
import threading
import jinja2
from past.builtins import basestring
from collections import defaultdict
from plynx.constants import JobReturnStatus, ParameterTypes
from plynx.db.node import Parameter, Output
from plynx.utils.file_handler import get_file_stream, upload_file_stream
from plynx.utils.config import get_worker_config
import plynx.utils.plugin_manager
from plynx.plugins.resources.common import FILE_KIND
import plynx.base.executor
from plynx.constants import NodeResources

WORKER_CONFIG = get_worker_config()


def _RESOURCE_MERGER_FUNC():
    return defaultdict(list)


class ResourceMerger(object):
    def __init__(self, init_level_0=None, init_level_1=None):
        self._dict = defaultdict(_RESOURCE_MERGER_FUNC)
        init_level_0 = init_level_0 or []
        init_level_1 = init_level_1 or []
        for key in init_level_0:
            self._dict[key] = _RESOURCE_MERGER_FUNC()
            for lev_1 in init_level_1:
                self._dict[key][lev_1] = []

    def append(self, resource_dict, resource_name, is_list):
        for key, value in resource_dict.items():
            if is_list:
                self._dict[key][resource_name].append(value)
            else:
                self._dict[key][resource_name] = value

    def get_dict(self):
        """
        Return original dict.

        .. highlight:: python
        .. code-block:: python
        .. {
        ..     'inputs': {
        ..         'input_name_0': [],
        ..         'input_name_1': ['/tmp/1', '/tmp/2'],
        ..         'input_name_2': '/tmp/3',
        ..     },
        ..     'cloud_inputs': {
        ..         'input_name_0': [],
        ..         'input_name_1': ['gs://1', 'gs://2'],
        ..         'input_name_2': 'gs://3',
        ..     }
        .. }
        ..

        Out:    Dict
        """
        return self._dict


class BaseBash(plynx.base.executor.BaseExecutor):

    def __init__(self, node=None):
        super(BaseBash, self).__init__(node)
        self.sp = None
        self.logs_sizes = {}
        self.final_logs_uploaded = False
        self.logs = {}
        self.logs_lock = threading.Lock()
        self._resource_manager = plynx.utils.plugin_manager.get_resource_manager()

    def exec_script(self, script_location, command='bash'):
        res = JobReturnStatus.SUCCESS

        try:
            def pre_exec():
                # Restore default signal disposition and invoke setsid
                for sig in ('SIGPIPE', 'SIGXFZ', 'SIGXFSZ'):
                    if hasattr(signal, sig):
                        signal.signal(getattr(signal, sig), signal.SIG_DFL)
                os.setsid()

            env = os.environ.copy()

            # append running script to worker log
            with open(script_location, 'r') as sf, open(self.logs['worker'], 'a') as wf:
                wf.write(self._make_debug_text("Running script:"))
                wf.write(sf.read())
                wf.write('\n')
                wf.write(self._make_debug_text("End script"))

            with open(self.logs['stdout'], 'wb') as stdout_file, open(self.logs['stderr'], 'wb') as stderr_file:
                self.sp = Popen(
                    [command, script_location],
                    stdout=stdout_file, stderr=stderr_file,
                    cwd=self.workdir, env=env,
                    preexec_fn=pre_exec)

                self.sp.wait()

            if self.sp.returncode:
                raise Exception("Process returned non-zero value")

        except Exception as e:
            res = JobReturnStatus.FAILED
            logging.exception("Job failed")
            with open(self.logs['worker'], 'a+') as worker_log_file:
                worker_log_file.write(self._make_debug_text("JOB FAILED"))
                worker_log_file.write(str(e))

        return res

    def kill(self):
        if not hasattr(self, 'sp') or not self.sp:
            return

        logging.info('Sending SIGTERM signal to bash process group')
        try:
            os.killpg(os.getpgid(self.sp.pid), signal.SIGTERM)
            logging.info('Killed {}'.format(self.sp.pid))
        except OSError as e:
            logging.error('Error: {}'.format(e))

    def is_updated(self):
        logging.info('Tick')
        return self.upload_logs(final=False)

    # Hack: do not pickle file
    def __getstate__(self):
        logging.critical('Run into `__getstate__`! Look ad `logs_lock` variable')
        d = dict(self.__dict__)
        if 'sp' in d:
            del d['sp']
        return d

    @staticmethod
    def _make_debug_text(text):
        content = '\n'.join(['# {}'.format(line) for line in text.split('\n')])
        return "{border}\n{content}\n{border}\n".format(
            border='#' * 40,
            content=content
        )

    @classmethod
    def get_default_node(cls, is_workflow):
        if is_workflow:
            raise Exception('This class cannot be a workflow')
        node = super().get_default_node(is_workflow)
        node.parameters.extend(
            [
                Parameter.from_dict({
                    'name': '_cmd',
                    'parameter_type': ParameterTypes.CODE,
                    'value': {
                        'mode': 'sh',
                        'value': 'set -e\n\n',
                    },
                    'mutable_type': False,
                    'publicable': False,
                    'removable': False,
                    }
                ),
                Parameter.from_dict({
                    'name': '_cacheable',
                    'parameter_type': ParameterTypes.BOOL,
                    'value': True,
                    'mutable_type': False,
                    'publicable': False,
                    'removable': False,
                }),
                Parameter.from_dict({
                    'name': '_timeout',
                    'parameter_type': ParameterTypes.INT,
                    'value': 600,
                    'mutable_type': False,
                    'publicable': True,
                    'removable': False
                }),
            ]
        )
        node.logs.extend(
            [
                Output.from_dict({
                    'name': 'stderr',
                    'file_type': FILE_KIND,
                }),
                Output({
                    'name': 'stdout',
                    'file_type': FILE_KIND,
                }),
                Output({
                    'name': 'worker',
                    'file_type': FILE_KIND,
                }),
            ]
        )
        return node

    def _prepare_inputs(self, preview=False):
        resource_merger = ResourceMerger(
            [NodeResources.INPUT],
            [input.name for input in self.node.inputs if input.is_array],
        )
        for input in self.node.inputs:
            if preview:
                for i, value in enumerate(range(input.min_count)):
                    filename = os.path.join(self.workdir, 'i_{}_{}'.format(i, input.name))
                    resource_merger.append(
                        self._resource_manager.kind_to_resource_class[input.file_type].prepare_input(filename, preview),
                        input.name,
                        input.is_array,
                    )
            else:
                for i, value in enumerate(input.values):
                    filename = os.path.join(self.workdir, 'i_{}_{}'.format(i, input.name))
                    with open(filename, 'wb') as f:
                        f.write(get_file_stream(value).read())
                    resource_merger.append(
                        self._resource_manager.kind_to_resource_class[input.file_type].prepare_input(filename, preview),
                        input.name,
                        input.is_array,
                    )
        return resource_merger.get_dict()

    def _prepare_outputs(self, preview=False):
        resource_merger = ResourceMerger(
            [NodeResources.OUTPUT],
            [output.name for output in self.node.outputs if output.is_array],
        )
        for output in self.node.outputs:
            filename = os.path.join(self.workdir, 'o_{}'.format(output.name))
            resource_merger.append(
                self._resource_manager.kind_to_resource_class[output.file_type].prepare_output(filename, preview),
                output.name,
                is_list=False
            )
        return resource_merger.get_dict()

    def _prepare_logs(self):
        with self.logs_lock:
            self.logs = {}
            for log in self.node.logs:
                filename = os.path.join(self.workdir, 'l_{}'.format(log.name))
                self.logs[log.name] = filename
                self.logs_sizes[log.name] = 0
            return self.logs

    def _get_script_fname(self, extension='.sh'):
        return os.path.join(self.workdir, "exec{}".format(extension))

    def _prepare_parameters(self):
        res = {}
        for parameter in self.node.parameters:
            value = None
            if parameter.parameter_type == ParameterTypes.ENUM:
                index = max(0, min(len(parameter.value.values) - 1, int(parameter.value.index)))
                value = parameter.value.values[index]
            elif parameter.parameter_type in [ParameterTypes.LIST_STR, ParameterTypes.LIST_INT]:
                if parameter.parameter_type == ParameterTypes.LIST_INT:
                    value = list(map(int, parameter.value))
                else:
                    value = parameter.value
            elif parameter.parameter_type == ParameterTypes.CODE:
                value = parameter.value.value
            elif parameter.parameter_type == ParameterTypes.INT:
                value = int(parameter.value)
            else:
                value = parameter.value
            res[parameter.name] = value
        return res

    def _postprocess_outputs(self, outputs):
        for key, filename in outputs.items():
            logging.info("Uploading output `{}` - `{}`".format(key, filename))
            if os.path.exists(filename):
                logging.info('path exists')
                matching_outputs = list(filter(lambda o: o.name == key, self.node.outputs))
                assert len(matching_outputs) == 1, "Found more that 1 output with the same name `{}`".format(key)
                filename = self._resource_manager.kind_to_resource_class[matching_outputs[0].file_type].postprocess_output(filename)
                logging.info(filename)
                with open(filename, 'rb') as f:
                    # resource_id
                    self.node.get_output_by_name(key).values = [upload_file_stream(f)]
                    logging.info(self.node.get_output_by_name(key).to_dict())
            else:
                raise IOError("Output `{}` (filename: `{}`) does not exist".format(key, filename))

    def _postprocess_logs(self):
        self.upload_logs(final=True)

    def _extract_cmd_text(self):
        parameter = self.node.get_parameter_by_name('_cmd')
        if parameter.parameter_type == ParameterTypes.CODE:
            return parameter.value.value
        elif parameter.parameter_type == ParameterTypes.TEXT:
            # backward compatibility: `cmd` used to have type of TEXT
            return parameter.value
        raise TypeError("Process returned non-zero value")

    def upload_logs(self, final=False):
        is_dirty = False
        with self.logs_lock:
            if self.final_logs_uploaded:
                return is_dirty
            self.final_logs_uploaded = final
            for key, filename in self.logs.items():
                if key not in self.logs_sizes:
                    # the logs have not been initialized yet
                    continue
                if os.path.exists(filename) and os.stat(filename).st_size != self.logs_sizes[key]:
                    is_dirty = True
                    log = self.node.get_log_by_name(key)
                    self.logs_sizes[key] = os.stat(filename).st_size
                    with open(filename, 'rb') as f:
                        # resource_id should be None if the file has not been uploaded yet
                        # otherwise assign it
                        log.values = [upload_file_stream(f, log.values[0] if len(log.values) > 0 else None)]
        return is_dirty


class BashJinja2(BaseBash):
    HELP_TEMPLATE = '''# Use templates: {}
# For example `{{{{ '{{{{' }}}} param['_timeout'] {{{{ '}}}}' }}}}` or `{{{{ '{{{{' }}}} input['abc'] {{{{ '}}}}' }}}}`

'''

    def __init__(self, node=None):
        super(BashJinja2, self).__init__(node)

    def run(self, preview=False):
        inputs = self._prepare_inputs(preview)
        parameters = self._prepare_parameters()
        outputs = self._prepare_outputs(preview)
        logs = self._prepare_logs()
        if preview:
            help = BashJinja2.HELP_TEMPLATE.format(list(inputs.keys()) + list(outputs.keys()) + [NodeResources.PARAM])
        else:
            help = ''
        cmd = '{help}{cmd}'.format(
            help=help,
            cmd=self._extract_cmd_text()
        )
        cmd_template = jinja2.Template(cmd)
        resources = inputs
        resources.update(outputs)
        cmd_string = cmd_template.render(
            params=parameters,
            logs=logs,
            **resources
        )
        if preview:
            return cmd_string

        script_location = self._get_script_fname()
        with open(script_location, 'w') as script_file:
            script_file.write(
                cmd_string
            )

        res = self.exec_script(script_location)

        self._postprocess_outputs(outputs[NodeResources.OUTPUT])
        self._postprocess_logs()

        return res

    def status(self):
        pass

    @classmethod
    def get_default_node(cls, is_workflow):
        node = super().get_default_node(is_workflow)
        node.title = 'New bash script'
        return node


class PythonNode(BaseBash):
    def __init__(self, node=None):
        super(PythonNode, self).__init__(node)

    def run(self, preview=False):
        inputs = self._prepare_inputs(preview)
        parameters = self._prepare_parameters()
        outputs = self._prepare_outputs(preview)
        logs = self._prepare_logs()
        cmd = self._extract_cmd_text()
        cmd_array = []
        cmd_array.extend([
            self._get_arguments_string(key, value)
            for key, value in inputs.items()
        ])
        cmd_array.extend([
            self._get_arguments_string(key, value)
            for key, value in outputs.items()
        ])
        cmd_array.extend([
            self._get_arguments_string(NodeResources.PARAM, parameters),
            self._get_arguments_string(NodeResources.LOG, logs),
            "\n",
            "# User code starts there:",
            cmd,
        ])
        cmd_string = '\n'.join(cmd_array)

        if preview:
            return cmd_string

        script_location = self._get_script_fname(extension='.py')
        with open(script_location, 'w') as script_file:
            script_file.write(
                cmd_string
            )

        res = self.exec_script(script_location, command='python')

        self._postprocess_outputs(outputs[NodeResources.OUTPUT])
        self._postprocess_logs()

        return res

    def status(self):
        """Temp"""
        pass

    @classmethod
    def _get_arguments_string(cls, var_name, arguments):
        res = ['{} = {{}}'.format(var_name)]
        for key, value in arguments.items():
            res.append('{var_name}["{key}"] = {value}'.format(
                var_name=var_name,
                key=key,
                value=repr(cls._pythonize(value))
                )
            )
        return '\n'.join(res)

    @staticmethod
    def _pythonize(value):
        if isinstance(value, basestring):
            repr(value)
        return value

    @classmethod
    def get_default_node(cls, is_workflow):
        node = super().get_default_node(is_workflow)
        node.title = 'New python script'
        param = list(filter(lambda p: p.name == '_cmd', node.parameters))[0]
        param.value.mode = 'python'
        param.value.value = 'print("hello world")'
        return node



from collections import namedtuple
from plynx.constants import NodeResources

PreviewObject = namedtuple('PreviewObject', ['fp', 'resource_id'])


def _force_decode(byte_array):
    try:
        return byte_array.decode('utf-8')
    except UnicodeDecodeError:
        return '# not a UTF-8 sequence:\n{}'.format(byte_array)
    return 'Failed to decode the sequence'


class BaseResource(object):
    def __init__(self):
        pass

    @staticmethod
    def prepare_input(filename, preview=False):
        return {NodeResources.INPUT: filename}

    @staticmethod
    def prepare_output(filename, preview=False):
        if not preview:
            with open(filename, 'a'):
                pass
        return {NodeResources.OUTPUT: filename}

    @staticmethod
    def postprocess_output(filename):
        return filename

    @classmethod
    def preview(cls, preview_object):
        preview_object.fp.truncate(1024 ** 2)  # 1 MB
        # TODO escape html code for security reasons
        return '<pre>{}</pre>'.format(_force_decode(preview_object.fp.read()))



#!/usr/bin/env python
from plynx.bin.cli import CLIFactory


def main():
    parser = CLIFactory.get_parser()
    args = parser.parse_args()
    kwargs = {
        k: v
        for k, v in vars(args).items() if k in args.args
    }
    CLIFactory.parse_global_config_parameters(kwargs)
    args.func(kwargs)



import plynx.db.node
import plynx.utils.exceptions
import plynx.utils.plugin_manager


def materialize_executor(node_dict):
    """
    Create a Node object from a dictionary

    Parameters:
    node_dict (dict): dictionary representation of a Node

    Returns:
    Node: Derived from plynx.nodes.base.Node class
    """
    kind = node_dict['kind']
    cls = plynx.utils.plugin_manager.get_executor_manager().kind_to_executor_class.get(kind)
    if not cls:
        raise plynx.utils.exceptions.NodeNotFound(
            'Node kind `{}` not found'.format(kind)
        )
    return cls(plynx.db.node.Node.from_dict(node_dict))



[
  {
    "_id": "5e66fc2df1e5b88d18f5c14f",
    "title": "Word2Vec 1",
    "_type": "Node",
    "description": "Find similar words",
    "kind": "basic-python-node-operation",
    "parent_node_id": null,
    "successor_node_id": null,
    "original_node_id": null,
    "inputs": [
      {
        "input_references": [
          {
            "node_id": "5e66fca5f1e5b88d18f5c23a",
            "output_id": "words"
          }
        ],
        "name": "positives",
        "file_type": "json",
        "values": [],
        "is_array": true,
        "min_count": 0
      },
      {
        "input_references": [
          {
            "node_id": "5e66fcc4f1e5b88d18f5c278",
            "output_id": "words"
          }
        ],
        "name": "negatives",
        "file_type": "json",
        "values": [],
        "is_array": true,
        "min_count": 0
      }
    ],
    "outputs": [
      {
        "name": "result",
        "file_type": "json",
        "values": [],
        "is_array": false,
        "min_count": 1
      }
    ],
    "parameters": [
      {
        "name": "_cmd",
        "parameter_type": "code",
        "value": {
          "value": "import json\nimport pickle\nimport os\n\n\ndef get_words(filenames):\n    res = []\n    for filename in filenames:\n        with open(filename) as f:\n            res.extend(json.load(f))\n    return res\n\npositive = get_words(inputs['positives'])\nnegative = get_words(inputs['negatives'])\n\nwith open(os.path.join('/models', params['model']), 'rb') as f:\n    model = pickle.load(f)\n\nres = model.most_similar(positive=positive, negative=negative)\n\nwith open(outputs['result'], 'w') as f:\n    json.dump(res, f)",
          "mode": "python"
        },
        "mutable_type": false,
        "removable": false,
        "publicable": false,
        "widget": null,
        "reference": null
      },
      {
        "name": "_cacheable",
        "parameter_type": "bool",
        "value": true,
        "mutable_type": false,
        "removable": false,
        "publicable": false,
        "widget": null,
        "reference": null
      },
      {
        "name": "_timeout",
        "parameter_type": "int",
        "value": 600,
        "mutable_type": false,
        "removable": false,
        "publicable": true,
        "widget": null,
        "reference": null
      },
      {
        "name": "model",
        "parameter_type": "enum",
        "value": {
          "values": [
            "glove-twitter-200",
            "glove-twitter-25"
          ],
          "index": "0"
        },
        "mutable_type": true,
        "removable": true,
        "publicable": true,
        "widget": "model",
        "reference": null
      }
    ],
    "logs": [
      {
        "name": "stderr",
        "file_type": "file",
        "values": [],
        "is_array": false,
        "min_count": 1
      },
      {
        "name": "stdout",
        "file_type": "file",
        "values": [],
        "is_array": false,
        "min_count": 1
      },
      {
        "name": "worker",
        "file_type": "file",
        "values": [],
        "is_array": false,
        "min_count": 1
      }
    ],
    "node_running_status": "CREATED",
    "node_status": "READY",
    "cache_url": "",
    "x": 282,
    "y": 142,
    "author": "5e5dd9c7653752f34e1e0be6",
    "starred": false
  },
  {
    "_id": "5e66fca5f1e5b88d18f5c33b",
    "title": "Words to list 1",
    "_type": "Node",
    "description": "Custom words",
    "kind": "basic-python-node-operation",
    "parent_node_id": null,
    "successor_node_id": null,
    "original_node_id": null,
    "inputs": [],
    "outputs": [
      {
        "name": "words",
        "file_type": "json",
        "values": [],
        "is_array": false,
        "min_count": 1
      }
    ],
    "parameters": [
      {
        "name": "_cmd",
        "parameter_type": "code",
        "value": {
          "value": "import json\n\nwith open(outputs['words'], 'w') as f:\n    json.dump(params['words'], f)",
          "mode": "python"
        },
        "mutable_type": false,
        "removable": false,
        "publicable": false,
        "widget": null,
        "reference": null
      },
      {
        "name": "_cacheable",
        "parameter_type": "bool",
        "value": true,
        "mutable_type": false,
        "removable": false,
        "publicable": false,
        "widget": null,
        "reference": null
      },
      {
        "name": "_timeout",
        "parameter_type": "int",
        "value": 600,
        "mutable_type": false,
        "removable": false,
        "publicable": true,
        "widget": null,
        "reference": null
      },
      {
        "name": "words",
        "parameter_type": "list_str",
        "value": [
          "king",
          "woman"
        ],
        "mutable_type": true,
        "removable": true,
        "publicable": true,
        "widget": "words",
        "reference": null
      }
    ],
    "logs": [
      {
        "name": "stderr",
        "file_type": "file",
        "values": [],
        "is_array": false,
        "min_count": 1
      },
      {
        "name": "stdout",
        "file_type": "file",
        "values": [],
        "is_array": false,
        "min_count": 1
      },
      {
        "name": "worker",
        "file_type": "file",
        "values": [],
        "is_array": false,
        "min_count": 1
      }
    ],
    "node_running_status": "CREATED",
    "node_status": "READY",
    "cache_url": "",
    "x": 30,
    "y": 80,
    "author": "5e5dd9c7653752f34e1e0be6",
    "starred": false
  },
  {
    "_id": "5e66fca5f1e5b88d18f5c43b",
    "title": "Words to list 2",
    "description": "Custom words",
    "kind": "_GROUP",
    "_type": "Group",
    "items": [
        {
          "_id": "5e66fca5f1e5b88d18f5c53b",
          "title": "Words to list 3",
          "description": "Custom words",
          "kind": "_GROUP",
          "_type": "Group",
          "items": [
              {"_id": "5e66fca5f1e5b88d18f5c23c","title": "Words to list 1","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c23d","title": "Words to list 2","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c23e","title": "Words to list 3","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c23f","title": "Words to list 4","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c24a","title": "Words to list 5","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c25a","title": "Words to list 6","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c26a","title": "Words to list 7","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c27a","title": "Words to list 8","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c28a","title": "Words to list 9","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c29a","title": "Words to list 10","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c2aa","title": "Words to list 11","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
              {"_id": "5e66fca5f1e5b88d18f5c2ba","title": "Words to list 12","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"}
          ],
          "author": "5e5dd9c7653752f34e1e0be6",
          "starred": false
        },
        {"_id": "5e66fca5f1e5b88d18f5c2ca","title": "Words to list a","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
        {"_id": "5e66fca5f1e5b88d18f5c2da","title": "Words to list b","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"},
        {"_id": "5e66fca5f1e5b88d18f5c2ea","title": "Words to list c","_type": "Node","description": "Custom words","kind": "basic-python-node-operation","parent_node_id": null,"successor_node_id": null,"original_node_id": null,"inputs": [],"outputs": [{"name": "words","file_type": "json","values": [],"is_array": false,"min_count": 1}],"parameters": [],"logs": [],"node_running_status": "CREATED","node_status": "READY","cache_url": "","author": "5e5dd9c7653752f34e1e0be6"}
    ],
    "author": "5e5dd9c7653752f34e1e0be6",
    "starred": false
  }
]



from plynx.db.db_object import DBObject, DBObjectField
from plynx.constants import Collections
from plynx.utils.common import ObjectId
from plynx.utils.db_connector import get_db_connector


class RunCancellation(DBObject):
    """RunCancellation represents Run Cancellation event in the database."""

    FIELDS = {
        '_id': DBObjectField(
            type=ObjectId,
            default=ObjectId,
            is_list=False,
            ),
        "run_id": DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
    }

    DB_COLLECTION = Collections.RUN_CANCELLATIONS


class RunCancellationManager(object):
    """RunCancellationManager contains basic operations related to `runs_cancellations` collection."""

    @staticmethod
    def cancel_run(run_id):
        """Cancel Run.
        Args:
            run_id    (ObjectId, str) RunID
        """
        run_cancellation = RunCancellation()
        run_cancellation.run_id = ObjectId(run_id)
        run_cancellation.save()
        return True

    @staticmethod
    def get_run_cancellations():
        """Get all Run Cancellation events"""
        res = []
        for runs_cancellation_dict in get_db_connector()[Collections.RUN_CANCELLATIONS].find():
            res.append(
                RunCancellation.from_dict(runs_cancellation_dict)
            )
        return res

    @staticmethod
    def remove(runs_cancellation_ids):
        """Remove Run Cancellation events with given Ids
        Args:
            runs_cancellation_ids     (list of ObjectID)  List of Run IDs to remove
        """
        get_db_connector()[Collections.RUN_CANCELLATIONS].delete_many({'_id': {'$in': runs_cancellation_ids}})



class ParameterTypes:
    STR = 'str'
    INT = 'int'
    BOOL = 'bool'
    TEXT = 'text'
    ENUM = 'enum'
    LIST_STR = 'list_str'
    LIST_INT = 'list_int'
    LIST_NODE = 'list_node'
    CODE = 'code'



from future.standard_library import install_aliases
install_aliases()   # noqa

import os                                                               # noqa: E402
import boto3                                                            # noqa: E402
from botocore.errorfactory import ClientError                           # noqa: E402
from urllib.parse import urlparse                                       # noqa: E402
from plynx.utils.remote.base import ContentsHandlerBase, RemoteBase     # noqa: E402


class ContentsHandlerS3(ContentsHandlerBase):
    def __init__(self, remote, path):
        super(ContentsHandlerS3, self).__init__(remote)
        self.path = urlparse(path).path[1:]     # get the file path in the bucket

    def get_contents_to_file(self, file_obj):
        self.remote.s3.download_fileobj(Bucket=self.remote.bucket_name, Key=self.path, Fileobj=file_obj)

    def set_contents_from_file(self, file_obj):
        self.remote.s3.upload_fileobj(Fileobj=file_obj, Bucket=self.remote.bucket_name, Key=self.path)

    def remove(self):
        self.remote.s3.delete_object(Bucket=self.remote.bucket_name, Key=self.path)

    def exists(self):
        try:
            self.remote.s3.head_object(Bucket=self.remote.bucket_name, Key=self.path)
        except ClientError:
            return False
        return True


class RemoteS3(RemoteBase):
    def __init__(self, storage_config):
        super(RemoteS3, self).__init__(ContentsHandlerS3, storage_config)
        if self._storage_config.credential_path:
            os.environ['AWS_SHARED_CREDENTIALS_FILE'] = self._storage_config.credential_path
        parse_result = urlparse(self._storage_config.prefix)
        self.s3 = boto3.client('s3')
        self.bucket_name = parse_result.netloc



from future.standard_library import install_aliases
install_aliases()   # noqa

import os                                                               # noqa: E402
from google.cloud import storage                                        # noqa: E402
from urllib.parse import urlparse                                       # noqa: E402
from plynx.utils.remote.base import ContentsHandlerBase, RemoteBase     # noqa: E402


class ContentsHandlerGS(ContentsHandlerBase):
    def __init__(self, remote, path):
        super(ContentsHandlerGS, self).__init__(remote)
        self.blob = self.remote.bucket.blob(urlparse(path).path[1:])

    def get_contents_to_file(self, file_obj):
        self.blob.download_to_file(file_obj)

    def set_contents_from_file(self, file_obj):
        self.blob.upload_from_file(file_obj)

    def remove(self):
        self.blob.delete()

    def exists(self):
        return self.blob.exists()


class RemoteGS(RemoteBase):
    def __init__(self, storage_config):
        super(RemoteGS, self).__init__(ContentsHandlerGS, storage_config)
        if self._storage_config.credential_path:
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = self._storage_config.credential_path
        bucket_name = urlparse(self._storage_config.prefix).netloc
        client = storage.Client()
        self.bucket = client.get_bucket(bucket_name)



import json
from plynx.utils import executor


def run_execute(filename):
    with open(filename) as f:
        node = executor.materialize_executor(json.load(f))
    node.run()
    print('Success')
    return 0



# flake8: noqa
from plynx.web.common import app, verify_password, authenticate, requires_auth, register_user, make_fail_response, run_api
import plynx.web.node
import plynx.web.resource
import plynx.web.user
import plynx.web.state
import plynx.web.health
from gevent.pywsgi import WSGIServer



from plynx.constants import Collections
from plynx.db.db_object import DBObject, DBObjectField
from plynx.utils.common import ObjectId


class Group(DBObject):
    """Group with db interface."""

    FIELDS = {
        '_id': DBObjectField(
            type=ObjectId,
            default=ObjectId,
            is_list=False,
            ),
        '_type': DBObjectField(
            type=str,
            default='Group',
            is_list=False,
            ),
        'title': DBObjectField(
            type=str,
            default='Title',
            is_list=False,
            ),
        # Kind, such as plynx.plugins.executors.local.BashJinja2. Derived from from plynx.plugins.executors.BaseExecutor class.
        'kind': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'items': DBObjectField(
            type=lambda x: x,   # Preserve type
            default=list,
            is_list=True,
            ),
        'author': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        'starred': DBObjectField(
            type=bool,
            default=False,
            is_list=False,
            ),
    }

    DB_COLLECTION = Collections.GROUPS

    def __str__(self):
        return 'Group(_id="{}")'.format(self._id)

    def __repr__(self):
        return 'Group({})'.format(str(self.to_dict()))



from bson.objectid import ObjectId


class JobReturnStatus:
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'


class NodeRunningStatus:
    STATIC = 'STATIC'
    CREATED = 'CREATED'
    READY = 'READY'
    IN_QUEUE = 'IN_QUEUE'
    RUNNING = 'RUNNING'
    SUCCESS = 'SUCCESS'
    RESTORED = 'RESTORED'
    FAILED = 'FAILED'
    FAILED_WAITING = 'FAILED_WAITING'
    CANCELED = 'CANCELED'
    SPECIAL = 'SPECIAL'

    _FAILED_STATUSES = {
        FAILED,
        CANCELED,
        FAILED_WAITING,
    }

    _SUCCEEDED_STATUSES = {
        STATIC,
        SUCCESS,
        RESTORED,
        SPECIAL,
    }

    _FINISHED_STATUSES = _FAILED_STATUSES | _SUCCEEDED_STATUSES

    @staticmethod
    def is_finished(node_running_status):
        return node_running_status in NodeRunningStatus._FINISHED_STATUSES

    @staticmethod
    def is_succeeded(node_running_status):
        return node_running_status in NodeRunningStatus._SUCCEEDED_STATUSES

    @staticmethod
    def is_failed(node_running_status):
        return node_running_status in NodeRunningStatus._FAILED_STATUSES


class NodeStatus:
    CREATED = 'CREATED'
    READY = 'READY'
    DEPRECATED = 'DEPRECATED'
    MANDATORY_DEPRECATED = 'MANDATORY_DEPRECATED'


class NodePostAction:
    SAVE = 'SAVE'
    APPROVE = 'APPROVE'
    CREATE_RUN = 'CREATE_RUN'
    CLONE = 'CLONE'
    VALIDATE = 'VALIDATE'
    DEPRECATE = 'DEPRECATE'
    MANDATORY_DEPRECATE = 'MANDATORY_DEPRECATE'
    PREVIEW_CMD = 'PREVIEW_CMD'
    REARRANGE_NODES = 'REARRANGE_NODES'
    UPGRADE_NODES = 'UPGRADE_NODES'
    CANCEL = 'CANCEL'
    GENERATE_CODE = 'GENERATE_CODE'


class NodePostStatus:
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'
    VALIDATION_FAILED = 'VALIDATION_FAILED'


class NodeClonePolicy:
    NODE_TO_NODE = 0
    NODE_TO_RUN = 1
    RUN_TO_NODE = 2


class NodeVirtualCollection:
    OPERATIONS = 'operations'
    WORKFLOWS = 'workflows'


class SpecialNodeId:
    INPUT = ObjectId('2419f9500000000000000000')
    OUTPUT = ObjectId('56274ccc0000000000000000')



import logging
import pymongo
from plynx.constants import Collections
from plynx.utils.config import get_db_config

_db = None


def init_indexes():
    _db[Collections.WORKER_STATE].create_index('insertion_date', expireAfterSeconds=5)

    _db[Collections.RUNS].create_index('insertion_date')

    _db[Collections.NODE_CACHE].create_index('key', unique=True)

    _db[Collections.TEMPLATES].create_index('insertion_date')
    _db[Collections.TEMPLATES].create_index([
        ('starred', pymongo.DESCENDING),
        ('insertion_date', pymongo.DESCENDING)
    ])
    _db[Collections.TEMPLATES].create_index([('title', pymongo.TEXT), ('description', pymongo.TEXT)])

    _db[Collections.RUNS].create_index('insertion_date')
    _db[Collections.RUNS].create_index([('title', pymongo.TEXT), ('description', pymongo.TEXT)])

    _db[Collections.USERS].create_index('username', unique=True)

    _db[Collections.RUN_CANCELLATIONS].create_index('insertion_date', expireAfterSeconds=60)
    _db[Collections.RUN_CANCELLATIONS].create_index('run_id')

    _db[Collections.GROUPS].create_index('insertion_date')
    _db[Collections.GROUPS].create_index([
        ('starred', pymongo.DESCENDING),
        ('insertion_date', pymongo.DESCENDING)
    ])
    _db[Collections.GROUPS].create_index([('title', pymongo.TEXT)])


def get_db_connector():
    global _db
    if _db:
        return _db
    connectionConfig = get_db_config()
    client = pymongo.MongoClient(connectionConfig.host, connectionConfig.port, read_preference=pymongo.read_preferences.PrimaryPreferred())
    _db = client['plynx']
    if connectionConfig.user:
        _db.authenticate(connectionConfig.user, connectionConfig.password)
    init_indexes()
    return _db


def check_connection():
    try:
        logging.info('Try db connection')
        get_db_connector().client.server_info()
    except pymongo.errors.ServerSelectionTimeoutError:
        logging.error('Connection failed')
        raise



import os
import sys
import json
import re
import collections
from datetime import datetime
from bson import ObjectId

SearchParameter = collections.namedtuple('SearchParameter', ['key', 'value'])

SEARCH_RGX = re.compile(r'[^\s]+:[^\s]+')

ObjectId = ObjectId


def to_object_id(_id):
    if type(_id) != ObjectId:
        _id = ObjectId(_id)
    return _id


class JSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ObjectId):
            return str(o)
        if isinstance(o, datetime):
            return str(o)
        return json.JSONEncoder.default(self, o)


def zipdir(path, zf):
    for root, dirs, files in os.walk(path):
        for file in files:
            arcname = os.path.relpath(os.path.join(root, file), os.path.join(path))
            zf.write(os.path.join(root, file), arcname)


def parse_search_string(search_string):
    found_matches = re.findall(SEARCH_RGX, search_string)
    search_parameters = dict([match.split(':') for match in found_matches])

    return search_parameters, ' '.join(re.sub(SEARCH_RGX, '', search_string).split())


def query_yes_no(question, default='yes'):
    """Ask a yes/no question via input() and return their answer.

    Args:
        question    (str):  String that is presented to the user.
        default     (str):  'yes' or 'no' default value

    The 'answer' return value is True for 'yes' or False for 'no'.
    """
    valid = {'yes': True, 'y': True, 'ye': True,
             'no': False, 'n': False}
    if default is None:
        prompt = ' [y/n] '
    elif default == 'yes':
        prompt = ' [Y/n] '
    elif default == 'no':
        prompt = ' [y/N] '
    else:
        raise ValueError('invalid default answer: `%s`' % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                             "(or 'y' or 'n').\n")


def update_dict_recursively(dest, u):
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            dest[k] = update_dict_recursively(dest.get(k, {}), v)
        else:
            dest[k] = v
    return dest



import os
import stat
import json
import zipfile
from plynx.constants import NodeResources
from plynx.base import resource
from plynx.utils.common import zipdir
from plynx.utils.config import get_web_config

WEB_CONFIG = get_web_config()


class File(resource.BaseResource):
    pass


class PDF(resource.BaseResource):
    @classmethod
    def preview(cls, preview_object):
        return '<iframe src="{}" title="preview" type="application/pdf" width="100%"/>'.format(
            '{}/resource/{}'.format(
                WEB_CONFIG.api_endpoint,
                preview_object.resource_id),
        )


class Image(resource.BaseResource):
    @classmethod
    def preview(cls, preview_object):
        return '<img src="{}" width="100%" alt="preview" />'.format(
            '{}/resource/{}'.format(
                WEB_CONFIG.api_endpoint,
                preview_object.resource_id),
        )


class _BaseSeparated(resource.BaseResource):
    SEPARATOR = None
    _ROW_CLASSES = ['even', 'odd']
    _NUM_ROW_CLASSES = len(_ROW_CLASSES)

    @classmethod
    def preview(cls, preview_object):
        preview_object.fp.truncate(1024 ** 2)
        line_data = []
        for idx, line in enumerate(preview_object.fp.read().decode('utf-8').split('\n')):
            line_data.append(
                '<tr class="preview-table-row-{}">{}</tr>'.format(
                    cls._ROW_CLASSES[idx % cls._NUM_ROW_CLASSES],
                    ''.join(map(lambda s: '<td class="preview-table-col">{}</td>'.format(s), line.split(cls.SEPARATOR)))
                )
            )
        return '<table class="preview-table">{}</table>'.format('\n'.join(line_data))


class CSV(_BaseSeparated):
    SEPARATOR = ','


class TSV(_BaseSeparated):
    SEPARATOR = '\t'


class Json(resource.BaseResource):
    @classmethod
    def preview(cls, preview_object):
        if preview_object.fp.getbuffer().nbytes > 1024 ** 2:
            return super(Json, cls).preview(preview_object)
        try:
            return '<pre>{}</pre>'.format(
                json.dumps(json.loads(preview_object.fp.read().decode('utf-8')), indent=2)
            )
        except Exception as e:
            return 'Failed to parse json: {}'.format(e)


class Executable(resource.BaseResource):
    @staticmethod
    def prepare_input(filename, preview):
        # `chmod +x` to the executable file
        if preview:
            return {NodeResources.INPUT: filename}
        st = os.stat(filename)
        os.chmod(filename, st.st_mode | stat.S_IEXEC)
        return {NodeResources.INPUT: filename}


class Directory(resource.BaseResource):
    @staticmethod
    def prepare_input(filename, preview):
        # extract zip file
        if preview:
            return {NodeResources.INPUT: filename}
        zip_filename = '{}.zip'.format(filename)
        os.rename(filename, zip_filename)
        os.mkdir(filename)
        with zipfile.ZipFile(zip_filename) as zf:
            zf.extractall(filename)
        return {NodeResources.INPUT: filename}

    @staticmethod
    def prepare_output(filename, preview):
        if preview:
            return {NodeResources.OUTPUT: filename}
        os.mkdir(filename)
        return {NodeResources.OUTPUT: filename}

    @staticmethod
    def postprocess_output(filename):
        zip_filename = '{}.zip'.format(filename)
        with zipfile.ZipFile(zip_filename, 'w', zipfile.ZIP_DEFLATED) as zf:
            zipdir(filename, zf)
        return zip_filename

    @classmethod
    def preview(cls, preview_object):
        with zipfile.ZipFile(preview_object.fp, 'r') as zf:
            content_stream = '\n'.join(zf.namelist())

        return '<pre>{}</pre>'.format(content_stream)


FILE_KIND = 'file'



import os
import shutil
from abc import abstractmethod
from plynx.db.node import Node, Parameter, ParameterTypes, NodeRunningStatus
from plynx.db.validation_error import ValidationError
from plynx.constants import NodeStatus, SpecialNodeId, ValidationTargetType, ValidationCode

TMP_DIR = '/tmp/plx'


class BaseExecutor:
    IS_GRAPH = False

    def __init__(self, node):
        self.node = node
        self.workdir = TMP_DIR

    @abstractmethod
    def run(self):
        """Main execution function.

        - Workdir has been initialized.
        - Inputs are not preprocessed.
        - Outputs shoul be manually postprocessed.
        - It is OK to raise an exception in this function.

        Returns:
            enum: plynx.constants.JobReturnStatus
        """
        pass

    @abstractmethod
    def status(self):
        """No currently used.
        """
        pass

    @abstractmethod
    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent executor canceled it.
        """
        pass

    def is_updated(self):
        """Function that is regularly called by a Worker.

        The function is running in a separate thread and does not block execution of `run()`.

        Returns:
            (bool):     True if worker needs to update DB else False
        """
        return False

    @classmethod
    def get_default_node(cls, is_workflow):
        node = Node()
        if cls.IS_GRAPH:
            nodes_parameter = Parameter.from_dict({
                'name': '_nodes',
                'parameter_type': ParameterTypes.LIST_NODE,
                'value': [],
                'mutable_type': False,
                'publicable': False,
                'removable': False,
                }
            )
            if not is_workflow:
                # need to add inputs and outputs
                import logging
                logging.info(type(nodes_parameter.value.value), 'a')
                nodes_parameter.value.value.extend(
                    [
                        Node.from_dict({
                            '_id': SpecialNodeId.INPUT,
                            'title': 'Input',
                            'kind': 'dummy',
                            'node_running_status': NodeRunningStatus.SPECIAL,
                            'node_status': NodeStatus.READY,
                        }),
                        Node.from_dict({
                            '_id': SpecialNodeId.OUTPUT,
                            'title': 'Output',
                            'kind': 'dummy',
                            'node_running_status': NodeRunningStatus.SPECIAL,
                            'node_status': NodeStatus.READY,
                        }),
                    ]
                )
            node.parameters.extend([
                nodes_parameter,
            ])
            node.arrange_auto_layout()
        return node

    def init_workdir(self):
        if not os.path.exists(self.workdir):
            os.makedirs(self.workdir)

    def clean_up(self):
        if os.path.exists(self.workdir):
            shutil.rmtree(self.workdir, ignore_errors=True)

    def validate(self):
        """Validate Node.

        Return:
            (ValidationError)   Validation error if found; else None
        """
        violations = []
        if self.node.title == '':
            violations.append(
                ValidationError(
                    target=ValidationTargetType.PROPERTY,
                    object_id='title',
                    validation_code=ValidationCode.MISSING_PARAMETER
                ))

        # Meaning the node is in the graph. Otherwise souldn't be in validation step
        if self.node.node_status != NodeStatus.CREATED:
            for input in self.node.inputs:
                min_count = input.min_count if input.is_array else 1
                if len(input.input_references) < min_count:
                    violations.append(
                        ValidationError(
                            target=ValidationTargetType.INPUT,
                            object_id=input.name,
                            validation_code=ValidationCode.MISSING_INPUT
                        ))

            if self.node.node_status == NodeStatus.MANDATORY_DEPRECATED:
                violations.append(
                    ValidationError(
                        target=ValidationTargetType.NODE,
                        object_id=str(self.node._id),
                        validation_code=ValidationCode.DEPRECATED_NODE
                    ))

        if len(violations) == 0:
            return None

        return ValidationError(
            target=ValidationTargetType.NODE,
            object_id=str(self.node._id),
            validation_code=ValidationCode.IN_DEPENDENTS,
            children=violations
        )


class Dummy(BaseExecutor):
    def __init__(self, node=None):
        super(Dummy, self).__init__(node)

    def run(self):
        raise NotImplementedError()

    def status(self):
        raise NotImplementedError()

    def kill(self):
        raise NotImplementedError()

    @classmethod
    def get_default_node(cls, is_workflow):
        raise NotImplementedError()



class ContentsHandlerBase(object):
    def __init__(self, remote):
        self.remote = remote

    def get_contents_to_file(self, file_obj):
        raise NotImplementedError()

    def set_contents_from_file(self, file_obj):
        raise NotImplementedError()

    def remove(self):
        raise NotImplementedError()

    def exists(self):
        raise NotImplementedError()


class RemoteBase(object):
    def __init__(self, ContentsHandler, storage_config):
        self.ContentsHandler = ContentsHandler
        self._storage_config = storage_config

    def get_contents_handler(self, path):
        full_path = '{}{}'.format(self._storage_config.prefix, path)
        return self.ContentsHandler(self, full_path)



import struct
import pickle


def send_msg(sock, msg):
    """Send message via socket"""
    # Prefix each message with a 4-byte length (network byte order)
    msg = pickle.dumps(msg)
    msg = struct.pack('>I', len(msg)) + msg
    sock.sendall(msg)


def recv_msg(sock):
    """Receive message via socket"""
    # Read message length and unpack it into an integer
    raw_msglen = _recvall(sock, 4)
    if not raw_msglen:
        return None
    msglen = struct.unpack('>I', raw_msglen)[0]
    # Read the message data
    msg = _recvall(sock, msglen)
    return pickle.loads(msg)


def _recvall(sock, n):
    """Helper function"""
    # Helper function to recv n bytes or return None if EOF is hit
    data = b''
    while len(data) < n:
        packet = sock.recv(n - len(data))
        if not packet:
            return None
        data += packet
    return data



from past.builtins import basestring
from collections import defaultdict, deque
from plynx.constants import Collections, NodeClonePolicy
from plynx.db.db_object import DBObject, DBObjectField
from plynx.utils.common import ObjectId
from plynx.constants import NodeStatus, NodeRunningStatus, SpecialNodeId
from plynx.plugins.resources.common import FILE_KIND
from plynx.constants import ParameterTypes


def _clone_update_in_place(node, node_clone_policy):
    if node.node_running_status == NodeRunningStatus.SPECIAL:
        for output in node.outputs:
            output.values = []
        return node
    old_node_id = node._id
    node._id = ObjectId()
    node.successor_node_id = None

    if node_clone_policy == NodeClonePolicy.NODE_TO_NODE:
        node.parent_node_id = old_node_id
        node.original_node_id = None
    elif node_clone_policy == NodeClonePolicy.NODE_TO_RUN:
        node.parent_node_id = None
        node.original_node_id = node.original_node_id or old_node_id
    elif node_clone_policy == NodeClonePolicy.RUN_TO_NODE:
        node.parent_node_id = node.original_node_id
        node.original_node_id = None
    else:
        raise Exception('Unknown clone policy `{}`'.format(node_clone_policy))

    if node.node_running_status == NodeRunningStatus.STATIC:
        return node
    node.node_running_status = NodeRunningStatus.READY
    node.node_status = NodeStatus.CREATED

    sub_nodes = node.get_parameter_by_name('_nodes', throw=False)
    if sub_nodes:
        object_id_mapping = {}
        for sub_node in sub_nodes.value.value:
            prev_id = ObjectId(sub_node._id)
            _clone_update_in_place(sub_node, node_clone_policy)
            object_id_mapping[prev_id] = sub_node._id

        for sub_node in sub_nodes.value.value:
            for input in sub_node.inputs:
                input.values = []
                for input_reference in input.input_references:
                    input_reference.node_id = object_id_mapping[ObjectId(input_reference.node_id)]
            # TODO STATIC
            for output in sub_node.outputs:
                output.values = []

            for log in sub_node.logs:
                log.values = []

            for parameter in sub_node.parameters:
                if not parameter.reference:
                    continue
                parameter.value = node.get_parameter_by_name(parameter.reference, throw=True).value

    for output_or_log in node.outputs + node.logs:
        output_or_log.resource_id = None
    return node


RESOURCE_FIELDS = {
    'name': DBObjectField(
        type=str,
        default='',
        is_list=False,
        ),
    'file_type': DBObjectField(
        type=str,
        default=FILE_KIND,
        is_list=False,
        ),
    'values': DBObjectField(
        type=lambda object_dict: object_dict,
        default=list,
        is_list=True,
        ),
    'is_array': DBObjectField(
        type=bool,
        default=False,
        is_list=False,
        ),
    'min_count': DBObjectField(
        type=int,
        default=1,
        is_list=False,
        ),
}


class Output(DBObject):
    """Basic Output structure."""

    FIELDS = RESOURCE_FIELDS

    def __str__(self):
        return 'Output(name="{}")'.format(self.name)

    def __repr__(self):
        return 'Output({})'.format(str(self.to_dict()))


class InputReference(DBObject):
    """Basic Value of the Input structure."""

    FIELDS = {
        'node_id': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'output_id': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
    }

    def __str__(self):
        return 'InputReference({}, {})'.format(self.node_id, self.output_id)

    def __repr__(self):
        return 'InputReference({})'.format(str(self.to_dict()))


class Input(DBObject):
    """Basic Input structure."""

    FIELDS = dict({
        'input_references': DBObjectField(
            type=InputReference,
            default=list,
            is_list=True,
            ),
        }, **RESOURCE_FIELDS)

    def __str__(self):
        return 'Input(name="{}")'.format(self.name)

    def __repr__(self):
        return 'Input({})'.format(str(self.to_dict()))


class Node(DBObject):
    """Basic Node with db interface."""

    FIELDS = {
        '_id': DBObjectField(
            type=ObjectId,
            default=ObjectId,
            is_list=False,
            ),
        '_type': DBObjectField(
            type=str,
            default='Node',
            is_list=False,
            ),
        'title': DBObjectField(
            type=str,
            default='Title',
            is_list=False,
            ),
        'description': DBObjectField(
            type=str,
            default='Description',
            is_list=False,
            ),
        # Kind, such as plynx.plugins.executors.local.BashJinja2. Derived from from plynx.plugins.executors.BaseExecutor class.
        'kind': DBObjectField(
            type=str,
            default='dummy',
            is_list=False,
            ),
        # ID of previous version of the node, always refer to `nodes` collection.
        'parent_node_id': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        # ID of next version of the node, always refer to `nodes` collection.
        'successor_node_id': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        # ID of original node, used in `runs`, always refer to `nodes` collection.
        # A Run refers to original node
        'original_node_id': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        'inputs': DBObjectField(
            type=Input,
            default=list,
            is_list=True,
            ),
        'outputs': DBObjectField(
            type=Output,
            default=list,
            is_list=True,
            ),
        'parameters': DBObjectField(
            type=lambda object_dict: Parameter(object_dict),
            default=list,
            is_list=True,
            ),
        'logs': DBObjectField(
            type=Output,
            default=list,
            is_list=True,
            ),
        'node_running_status': DBObjectField(
            type=str,
            default=NodeRunningStatus.CREATED,
            is_list=False,
            ),
        'node_status': DBObjectField(
            type=str,
            default=NodeStatus.CREATED,
            is_list=False,
            ),
        'cache_url': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'x': DBObjectField(
            type=int,
            default=0,
            is_list=False,
            ),
        'y': DBObjectField(
            type=int,
            default=0,
            is_list=False,
            ),
        'author': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        'starred': DBObjectField(
            type=bool,
            default=False,
            is_list=False,
            ),
    }

    DB_COLLECTION = Collections.TEMPLATES

    def _DEFAULT_LOG(name):
        return Output.from_dict({
            'name': name,
            'file_type': FILE_KIND,
            'values': [],
            'is_array': False,
            'min_count': 1,
        })

    def apply_properties(self, other_node):
        """Apply Properties and Inputs of another Node.
        This method is used for updating nodes.

        Args:
            other_node  (Node):     A node to copy Properties and Inputs from
        """
        for other_input in other_node.inputs:
            for input in self.inputs:
                if other_input.name == input.name and \
                   other_input.file_type == input.file_type and (
                            input.is_array or
                            (not input.is_array and 1 == len(other_input.input_references))
                       ):
                    input.input_references = other_input.input_references
                    break

        for other_parameter in other_node.parameters:
            for parameter in self.parameters:
                if other_parameter.name == parameter.name:
                    if parameter.parameter_type == other_parameter.parameter_type and parameter.widget:
                        parameter.value = other_parameter.value
                    break

        self.description = other_node.description

        self.x = other_node.x
        self.y = other_node.y

    def clone(self, node_clone_policy):
        return _clone_update_in_place(self.copy(), node_clone_policy)

    def __str__(self):
        return 'Node(_id="{}")'.format(self._id)

    def __repr__(self):
        return 'Node({})'.format(str(self.to_dict()))

    def _get_custom_element(self, arr, name, throw, default=None):
        for parameter in arr:
            if parameter.name == name:
                return parameter
        if throw:
            raise Exception('Parameter "{}" not found in {}'.format(name, self.title))
        if default:
            arr.append(default(name))
            return arr[-1]
        return None

    def get_input_by_name(self, name, throw=True):
        return self._get_custom_element(self.inputs, name, throw)

    def get_parameter_by_name(self, name, throw=True):
        return self._get_custom_element(self.parameters, name, throw)

    def get_output_by_name(self, name, throw=True):
        return self._get_custom_element(self.outputs, name, throw)

    def get_log_by_name(self, name, throw=False):
        return self._get_custom_element(self.logs, name, throw, default=Node._DEFAULT_LOG)

    def arrange_auto_layout(self, readonly=False):
        """Use heuristic to rearange nodes."""
        HEADER_HEIGHT = 23
        TITLE_HEIGHT = 20
        FOOTER_HEIGHT = 10
        BORDERS_HEIGHT = 2
        ITEM_HEIGHT = 20
        SPACE_HEIGHT = 50
        LEFT_PADDING = 30
        TOP_PADDING = 80
        LEVEL_WIDTH = 252
        SPECIAL_PARAMETER_HEIGHT = 20
        SPECIAL_PARAMETER_TYPES = [ParameterTypes.CODE]
        min_node_height = HEADER_HEIGHT + TITLE_HEIGHT + FOOTER_HEIGHT + BORDERS_HEIGHT

        node_id_to_level = defaultdict(lambda: -1)
        node_id_to_node = {}
        queued_node_ids = set()
        children_ids = defaultdict(set)

        sub_nodes = self.get_parameter_by_name('_nodes').value.value

        if len(sub_nodes) == 0:
            return

        node_ids = set([node._id for node in sub_nodes])
        non_zero_node_ids = set()
        for node in sub_nodes:
            node_id_to_node[node._id] = node
            for input in node.inputs:
                for input_reference in input.input_references:
                    parent_node_id = ObjectId(input_reference.node_id)
                    non_zero_node_ids.add(parent_node_id)
                    children_ids[parent_node_id].add(node._id)

        leaves = node_ids - non_zero_node_ids
        to_visit = deque()
        # Alwasy put Output Node in the end
        push_special = True if SpecialNodeId.OUTPUT in leaves and len(leaves) > 1 else False
        for leaf_id in leaves:
            node_id_to_level[leaf_id] = 1 if push_special and leaf_id != SpecialNodeId.OUTPUT else 0
            to_visit.append(leaf_id)

        while to_visit:
            node_id = to_visit.popleft()
            node = node_id_to_node[node_id]
            node_level = max([node_id_to_level[node_id]] + [node_id_to_level[child_id] + 1 for child_id in children_ids[node_id]])
            node_id_to_level[node_id] = node_level
            for input in node.inputs:
                for input_reference in input.input_references:
                    parent_node_id = ObjectId(input_reference.node_id)
                    parent_level = node_id_to_level[parent_node_id]
                    node_id_to_level[parent_node_id] = max(node_level + 1, parent_level)
                    if parent_node_id not in queued_node_ids:
                        to_visit.append(parent_node_id)
                        queued_node_ids.add(parent_node_id)

        max_level = max(node_id_to_level.values())
        level_to_node_ids = defaultdict(list)
        row_heights = defaultdict(lambda: 0)

        def get_index_helper(node, level):
            if level < 0:
                return 0
            parent_node_ids = set()
            for input in node.inputs:
                for input_reference in input.input_references:
                    parent_node_ids.add(ObjectId(input_reference.node_id))

            for index, node_id in enumerate(level_to_node_ids[level]):
                if node_id in parent_node_ids:
                    return index
            return -1

        def get_index(node, max_level, level):
            return tuple(
                [get_index_helper(node, lvl) for lvl in range(max_level, level, -1)]
            )

        for node_id, level in node_id_to_level.items():
            level_to_node_ids[level].append(node_id)

        # Push Input Node up the level
        if SpecialNodeId.INPUT in node_id_to_level and \
                (node_id_to_level[SpecialNodeId.INPUT] != max_level or len(level_to_node_ids[max_level]) > 1):
            input_level = node_id_to_level[SpecialNodeId.INPUT]
            level_to_node_ids[input_level] = [node_id for node_id in level_to_node_ids[input_level] if node_id != SpecialNodeId.INPUT]
            max_level += 1
            node_id_to_level[SpecialNodeId.INPUT] = max_level
            level_to_node_ids[max_level] = [SpecialNodeId.INPUT]

        for level in range(max_level, -1, -1):
            level_node_ids = level_to_node_ids[level]
            index_to_node_id = []
            for node_id in level_node_ids:
                node = node_id_to_node[node_id]
                index = get_index(node, max_level, level)
                index_to_node_id.append((index, node_id))

            index_to_node_id.sort()
            level_to_node_ids[level] = [node_id for _, node_id in index_to_node_id]

            for index, node_id in enumerate(level_to_node_ids[level]):
                node = node_id_to_node[node_id]
                special_parameters_count = sum(
                    1 if parameter.parameter_type in SPECIAL_PARAMETER_TYPES and parameter.widget else 0
                    for parameter in node.parameters
                )
                node_height = sum([
                    min_node_height,
                    ITEM_HEIGHT * max(len(node.inputs), len(node.outputs)),
                    special_parameters_count * SPECIAL_PARAMETER_HEIGHT
                ])
                row_heights[index] = max(row_heights[index], node_height)

        # TODO compute grid in a separate function
        if readonly:
            return level_to_node_ids, node_id_to_node

        cum_heights = [0]
        for index in range(len(row_heights)):
            cum_heights.append(cum_heights[-1] + row_heights[index] + SPACE_HEIGHT)

        max_height = max(cum_heights)

        for level in range(max_level, -1, -1):
            level_node_ids = level_to_node_ids[level]
            level_height = cum_heights[len(level_node_ids)]
            level_padding = (max_height - level_height) // 2
            for index, node_id in enumerate(level_node_ids):
                node = node_id_to_node[node_id]
                node.x = LEFT_PADDING + (max_level - level) * LEVEL_WIDTH
                node.y = TOP_PADDING + level_padding + cum_heights[index]


class ParameterEnum(DBObject):
    """Enum value."""

    FIELDS = {
        'values': DBObjectField(
            type=str,
            default=list,
            is_list=True,
            ),
        'index': DBObjectField(
            type=str,
            default=-1,
            is_list=False,
            ),
    }

    def __repr__(self):
        return 'ParameterEnum({})'.format(str(self.to_dict()))


class ParameterCode(DBObject):
    """Code value."""

    FIELDS = {
        'value': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'mode': DBObjectField(
            type=str,
            default='python',
            is_list=False,
            ),
    }

    def __repr__(self):
        return 'ParameterCode({})'.format(str(self.to_dict()))


class ParameterListOfNodes(DBObject):
    """List Of Nodes value."""

    FIELDS = {
        'value': DBObjectField(
            type=Node,
            default=list,
            is_list=True,
            ),
    }

    def __repr__(self):
        return 'ParameterListOfNodes({})'.format(str(self.to_dict()))


def _get_default_by_type(parameter_type):
    if parameter_type == ParameterTypes.STR:
        return ''
    if parameter_type == ParameterTypes.INT:
        return 0
    if parameter_type == ParameterTypes.BOOL:
        return False
    if parameter_type == ParameterTypes.TEXT:
        return ''
    if parameter_type == ParameterTypes.ENUM:
        return ParameterEnum()
    if parameter_type == ParameterTypes.LIST_STR:
        return []
    if parameter_type == ParameterTypes.LIST_INT:
        return []
    elif parameter_type == ParameterTypes.LIST_NODE:
        return ParameterListOfNodes()
    elif parameter_type == ParameterTypes.CODE:
        return ParameterCode()
    else:
        return None


def _value_is_valid(value, parameter_type):
    if parameter_type == ParameterTypes.STR:
        return isinstance(value, basestring)
    if parameter_type == ParameterTypes.INT:
        try:
            int(value)
        except Exception:
            return False
        return True
    if parameter_type == ParameterTypes.BOOL:
        return isinstance(value, int)
    if parameter_type == ParameterTypes.TEXT:
        return isinstance(value, basestring)
    if parameter_type == ParameterTypes.ENUM:
        return isinstance(value, ParameterEnum)
    if parameter_type == ParameterTypes.LIST_STR:
        return isinstance(value, list) and all(_value_is_valid(x, ParameterTypes.STR) for x in value)
    if parameter_type == ParameterTypes.LIST_INT:
        return isinstance(value, list) and all(_value_is_valid(x, ParameterTypes.INT) for x in value)
    elif parameter_type == ParameterTypes.LIST_NODE:
        # TODO proper validation
        return isinstance(value, ParameterListOfNodes)
    if parameter_type == ParameterTypes.CODE:
        return isinstance(value, ParameterCode)
    else:
        return False


class Parameter(DBObject):
    """Basic Parameter structure."""

    FIELDS = {
        'name': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'parameter_type': DBObjectField(
            type=str,
            default=ParameterTypes.STR,
            is_list=False,
            ),
        # TODO make type factory
        'value': DBObjectField(
            type=lambda x: x,   # Preserve type
            default='',
            is_list=False,
            ),
        'mutable_type': DBObjectField(
            type=bool,
            default=True,
            is_list=False,
            ),
        'removable': DBObjectField(
            type=bool,
            default=True,
            is_list=False,
            ),
        'publicable': DBObjectField(
            type=bool,
            default=True,
            is_list=False,
            ),
        'widget': DBObjectField(
            type=str,
            default=None,
            is_list=False,
            ),
        # Link to global parameter
        'reference': DBObjectField(
            type=str,
            default=None,
            is_list=False,
            ),
    }

    def __init__(self, obj_dict=None):
        super(Parameter, self).__init__(obj_dict)

        # `value` field is a special case: the type depends on `parameter_type`
        if self.value is None:
            self.value = _get_default_by_type(self.parameter_type)
        elif self.parameter_type == ParameterTypes.ENUM:
            self.value = ParameterEnum.from_dict(self.value)
        elif self.parameter_type == ParameterTypes.CODE:
            self.value = ParameterCode.from_dict(self.value)
        elif self.parameter_type == ParameterTypes.LIST_NODE:
            self.value = ParameterListOfNodes.from_dict(self.value)
        if not _value_is_valid(self.value, self.parameter_type):
            raise ValueError("Invalid parameter value type: {}: {}".format(self.name, self.value))

    def __str__(self):
        return 'Parameter(name="{}")'.format(self.name)

    def __repr__(self):
        return 'Parameter({})'.format(str(self.to_dict()))



from collections import namedtuple
from abc import abstractmethod

_QUERY_FIELDS = (
    ('status', ''),
    ('search', ''),
    ('per_page', 30),
    ('offset', 0),
    ('user_id', None),
)

Query = namedtuple(
    'Query',
    field_names=[field[0] for field in _QUERY_FIELDS],
    defaults=[field[1] for field in _QUERY_FIELDS],
    )


class BaseHub(object):
    def __init__(self):
        pass

    @abstractmethod
    def search(self, query):
        pass



class ValidationTargetType:
    BLOCK = 'BLOCK'
    GRAPH = 'GRAPH'
    INPUT = 'INPUT'
    NODE = 'NODE'
    PARAMETER = 'PARAMETER'
    PROPERTY = 'PROPERTY'


class ValidationCode:
    IN_DEPENDENTS = 'IN_DEPENDENTS'
    MISSING_INPUT = 'MISSING_INPUT'
    MISSING_PARAMETER = 'MISSING_PARAMETER'
    INVALID_VALUE = 'INVALID_VALUE'
    DEPRECATED_NODE = 'DEPRECATED_NODE'
    EMPTY_GRAPH = 'EMPTY_GRAPH'



from plynx.db.user import User


LIST_USERS = 'list_users'
CREATE_USER = 'create_user'
ACTIVATE_USER = 'activate_user'
DEACTIVATE_USER = 'deactivate_user'
MODES = [
    LIST_USERS,
    CREATE_USER,
    ACTIVATE_USER,
    DEACTIVATE_USER,
]


def run_list_users():
    for user_dict in User.find_users():
        user = User.from_dict(user_dict)
        print(','.join(map(str, [user._id, user.username])))


def run_create_user(username, password):
    if not username:
        raise ValueError('Usernmane must be specified')
    password = password or ''
    user = User()
    user.username = username
    user.hash_password(password)
    user.save()
    print('User `{}` created'.format(username))


def run_set_activation(username, value):
    user = User.find_user_by_name(username)

    if not user:
        raise ValueError("Username `{}` not found".format(username))

    user.active = value
    user.save()
    print('User`s `{}` active state changed to {}'.format(username, value))


def run_users(mode, username=None, password=''):
    if mode not in MODES:
        raise ValueError('`mode` must be one of `{values}`. Value `{mode}` is given'.format(
            values=MODES,
            mode=mode,
        ))
    if mode == LIST_USERS:
        run_list_users()
    elif mode == CREATE_USER:
        run_create_user(username, password)
    elif mode == ACTIVATE_USER:
        run_set_activation(username, True)
    elif mode == DEACTIVATE_USER:
        run_set_activation(username, False)



import os
import sys
import threading
import logging
import six
import traceback
import uuid
import socket
from plynx.constants import JobReturnStatus, NodeRunningStatus, Collections
import plynx.db.node_collection_manager
import plynx.db.run_cancellation_manager
from plynx.db.worker_state import WorkerState
from plynx.utils.config import get_worker_config
from plynx.utils.db_connector import check_connection
import plynx.utils.executor
from plynx.utils.file_handler import upload_file_stream


class TickThread(object):
    """
    This class is a Context Manager wrapper.
    It calls method `tick()` of the executor with a given interval
    """

    TICK_TIMEOUT = 1

    def __init__(self, executor):
        self.executor = executor
        self._tick_thread = threading.Thread(target=self.call_executor_tick, args=())
        self._stop_event = threading.Event()

    def __enter__(self):
        """
        Currently no meaning of returned class
        """
        self._tick_thread.start()
        return self

    def __exit__(self, type, value, traceback):
        self._stop_event.set()

    def call_executor_tick(self):
        while not self._stop_event.is_set():
            self._stop_event.wait(timeout=TickThread.TICK_TIMEOUT)
            if self.executor.is_updated():
                # Save logs whe operation is running
                self.executor.node.save(collection=Collections.RUNS)


class Worker(object):
    """Worker main class.

    Args:
        server_address      (tuple):        Define the server address (host, port).
        RequestHandlerClass (TCP handler):  Class of socketserver.BaseRequestHandler.

    Currently implemented as a TCP server.

    Only a single instance of the Worker class in a cluster is allowed.

    On the high level Worker distributes Jobs over all available Workers
    and updates statuses of the Graphs in the database.

    Worker performs several roles:
        * Pull graphs in status READY from the database.
        * Create Schedulers for each Graph.
        * Populate the queue of the Jobs.
        * Distribute Jobs accross Workers.
        * Keep track of Job's statuses.
        * Process CANCEL requests.
        * Update Graph's statuses.
        * Track worker status and last response.
    """

    # Define sync with database timeout
    SDB_STATUS_UPDATE_TIMEOUT = 1

    # Worker State update timeout
    WORKER_STATE_UPDATE_TIMEOUT = 1

    def __init__(self, worker_config, worker_id):
        self.worker_id = worker_id if worker_id else str(uuid.uuid1())
        self.node_collection_manager = plynx.db.node_collection_manager.NodeCollectionManager(collection=Collections.RUNS)
        self.run_cancellation_manager = plynx.db.run_cancellation_manager.RunCancellationManager()
        self.kinds = worker_config.kinds
        self.host = socket.gethostname()
        self._stop_event = threading.Event()

        # Mapping keep track of Worker Status
        self._run_id_to_executor = {}
        self._run_id_to_executor_lock = threading.Lock()

        # Start new threads
        self._thread_db_status_update = threading.Thread(target=self._run_db_status_update, args=())
        self._thread_db_status_update.start()

        self._thread_worker_state_update = threading.Thread(target=self._run_worker_state_update, args=())
        self._thread_worker_state_update.start()

        self._killed_run_ids = set()

    def serve_forever(self):
        """
        Run the worker.
        """
        self._stop_event.wait()

    def execute_job(self, executor):
        try:
            try:
                status = JobReturnStatus.FAILED
                executor.workdir = os.path.join('/tmp', str(uuid.uuid1()))
                executor.init_workdir()
                with TickThread(executor):
                    status = executor.run()
            except Exception:
                try:
                    f = six.BytesIO()
                    f.write(traceback.format_exc().encode())
                    executor.node.get_log_by_name('worker').resource_id = upload_file_stream(f)
                    logging.error(traceback.format_exc())
                except Exception:
                    # This case of `except` has happened before due to I/O failure
                    logging.critical(traceback.format_exc())
                    raise
            finally:
                executor.clean_up()

            logging.info('Node {node_id} `{title}` finished with status `{status}`'.format(
                node_id=executor.node._id,
                title=executor.node.title,
                status=status,
                ))
            if status == JobReturnStatus.SUCCESS:
                executor.node.node_running_status = NodeRunningStatus.SUCCESS
            elif executor.node._id in self._killed_run_ids:
                self._killed_run_ids.remove(executor.node._id)
                executor.node.node_running_status = NodeRunningStatus.CANCELED
            elif status == JobReturnStatus.FAILED:
                executor.node.node_running_status = NodeRunningStatus.FAILED
            else:
                raise Exception("Unknown return status value: `{}`".format(status))
        except Exception as e:
            logging.warning('Execution failed: {}'.format(e))
            executor.node.node_running_status = NodeRunningStatus.FAILED
        finally:
            executor.node.save(collection=Collections.RUNS)
            with self._run_id_to_executor_lock:
                del self._run_id_to_executor[executor.node._id]

    def _run_db_status_update(self):
        """Syncing with the database."""
        try:
            while not self._stop_event.is_set():
                node = self.node_collection_manager.pick_node(kinds=self.kinds)
                if node:
                    logging.info('New node found: {} {} {}'.format(node['_id'], node['node_running_status'], node['title']))
                    executor = plynx.utils.executor.materialize_executor(node)

                    with self._run_id_to_executor_lock:
                        self._run_id_to_executor[executor.node._id] = executor
                    thread = threading.Thread(target=self.execute_job, args=(executor, ))
                    thread.start()

                else:
                    self._stop_event.wait(timeout=Worker.SDB_STATUS_UPDATE_TIMEOUT)
        except Exception:
            self.stop()
            raise
        finally:
            logging.info("Exit {}".format(self._run_db_status_update.__name__))

    def _run_worker_state_update(self):
        """Syncing with the database."""
        try:
            while not self._stop_event.is_set():
                # TODO move CANCEL to a separate thread
                run_cancellations = list(self.run_cancellation_manager.get_run_cancellations())
                run_cancellation_ids = set(map(lambda rc: rc.run_id, run_cancellations)) & set(self._run_id_to_executor.keys())
                for run_id in run_cancellation_ids:
                    self._killed_run_ids.add(run_id)
                    self._run_id_to_executor[run_id].kill()

                runs = []
                with self._run_id_to_executor_lock:
                    for executor in self._run_id_to_executor.values():
                        runs.append(executor.node.to_dict())
                worker_state = WorkerState.from_dict({
                    'worker_id': self.worker_id,
                    'host': self.host,
                    'runs': runs,
                    'kinds': self.kinds,
                })
                worker_state.save()
                self._stop_event.wait(timeout=Worker.WORKER_STATE_UPDATE_TIMEOUT)
        except Exception:
            self.stop()
            raise
        finally:
            logging.info("Exit {}".format(self._run_worker_state_update.__name__))

    def stop(self):
        """Stop worker."""
        self._stop_event.set()


def run_worker(worker_id=None):
    """Run worker daemon. It will run in the same thread."""
    # Check connection to the db
    check_connection()

    logging.info('Init Worker')
    worker_config = get_worker_config()
    logging.info(worker_config)
    worker = Worker(worker_config, worker_id)

    # Activate the server; this will keep running until you
    # interrupt the program with Ctrl-C
    try:
        logging.info("Start serving")
        worker.serve_forever()
    except KeyboardInterrupt:
        worker.stop()
        sys.exit(0)



import hashlib
from builtins import str
from plynx.constants import Collections
from plynx.db.db_object import DBObject, DBObjectField
from plynx.db.node import Output
from plynx.utils.common import ObjectId
from plynx.utils.config import get_demo_config

demo_config = get_demo_config()


class NodeCache(DBObject):
    """Basic Node Cache with db interface."""

    FIELDS = {
        '_id': DBObjectField(
            type=ObjectId,
            default=ObjectId,
            is_list=False,
            ),
        'key': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'graph_id': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        'node_id': DBObjectField(
            type=ObjectId,
            default=None,
            is_list=False,
            ),
        'outputs': DBObjectField(
            type=Output,
            default=list,
            is_list=True,
            ),
        'logs': DBObjectField(
            type=Output,
            default=list,
            is_list=True,
            ),
        # `protected` is used to prevent removing saved cache
        'protected': DBObjectField(
            type=bool,
            default=False,
            is_list=False,
            ),
        'removed': DBObjectField(
            type=bool,
            default=False,
            is_list=False,
            ),
    }

    DB_COLLECTION = Collections.NODE_CACHE

    IGNORED_PARAMETERS = {'cmd', '_timeout'}

    @staticmethod
    def instantiate(node, graph_id, user_id):
        """Instantiate a Node Cache from Node.

        Args:
            node        (Node):             Node object
            graph_id    (ObjectId, str):    Graph ID
            user_id     (ObjectId, str):    User ID

        Return:
            (NodeCache)
        """

        return NodeCache({
            'key': NodeCache.generate_key(node, user_id),
            'node_id': node._id,
            'graph_id': graph_id,
            'outputs': [output.to_dict() for output in node.outputs],
            'logs': [log.to_dict() for log in node.logs],
        })

    # TODO after Demo: remove user_id
    @staticmethod
    def generate_key(node, user_id):
        """Generate hash.

        Args:
            node        (Node):             Node object
            user_id     (ObjectId, str):    User ID

        Return:
            (str)   Hash value
        """
        if not demo_config.enabled:
            user_id = ''    # TODO after demo
        inputs = node.inputs
        parameters = node.parameters
        original_node_id = node.original_node_id

        sorted_inputs = sorted(inputs, key=lambda x: x.name)
        inputs_hash = ','.join([
            '{}:{}'.format(
                input.name,
                ','.join(sorted(map(lambda x: x.resource_id, input.values)))
            )
            for input in sorted_inputs
        ])

        sorted_parameters = sorted(parameters, key=lambda x: x.name)
        parameters_hash = ','.join([
            '{}:{}'.format(
                parameter.name,
                parameter.value
            )
            for parameter in sorted_parameters if parameter.name not in NodeCache.IGNORED_PARAMETERS
        ])

        return hashlib.sha256(
            '{};{};{};{}'.format(
                original_node_id,
                inputs_hash,
                parameters_hash,
                str(user_id)).encode('utf-8')
        ).hexdigest()

    def __str__(self):
        return 'NodeCache(_id="{}")'.format(self._id)

    def __repr__(self):
        return 'NodeCache({})'.format(str(self.to_dict()))



class Collections:
    GROUPS = 'groups'
    NODE_CACHE = 'node_cache'
    RUN_CANCELLATIONS = 'run_cancellations'
    RUNS = 'runs'
    TEMPLATES = 'templates'
    USERS = 'users'
    WORKER_STATE = 'worker_state'



from plynx.constants import Collections
from plynx.db.db_object import DBObject, DBObjectField
from plynx.db.node import Node
from plynx.utils.db_connector import get_db_connector
from plynx.utils.common import ObjectId


class WorkerState(DBObject):
    """Worker statistics snapshot."""

    FIELDS = {
        '_id': DBObjectField(
            type=ObjectId,
            default=ObjectId,
            is_list=False,
            ),
        'worker_id': DBObjectField(
            type=str,
            default=None,
            is_list=False,
            ),
        'host': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'runs': DBObjectField(
            type=Node,
            default=list,
            is_list=True,
            ),
        'kinds': DBObjectField(
            type=str,
            default=list,
            is_list=True,
            ),
    }

    DB_COLLECTION = Collections.WORKER_STATE


def get_worker_states():
    states = getattr(get_db_connector(), Collections.WORKER_STATE)\
           .find({}).sort('insertion_date', -1)

    unique_worker_states = {}
    for state in states:
        if state['worker_id'] in unique_worker_states:
            continue
        unique_worker_states[state['worker_id']] = state

    return list(sorted(unique_worker_states.values(), key=lambda state: state['worker_id']))



from plynx.base import hub
from plynx.db.node_collection_manager import NodeCollectionManager


class CollectionHub(hub.BaseHub):
    def __init__(self, collection, operations):
        super(CollectionHub, self).__init__()

        self.node_collection_manager = NodeCollectionManager(collection=collection)
        self.operations = operations

    def search(self, query):
        return self.node_collection_manager.get_db_objects(node_kinds=self.operations, **query._asdict())



from plynx.utils.config import get_storage_config

_driver = None


def get_driver(storage_config=None):
    global _driver
    if not _driver:
        _driver = _get_driver_handler(storage_config)
    return _driver


def _get_driver_handler(storage_config=None):
    if storage_config is None:
        storage_config = get_storage_config()
    if storage_config.scheme == 'file':
        from plynx.utils.remote.file import RemoteFile
        return RemoteFile(storage_config)
    if storage_config.scheme == 'gs':
        from plynx.utils.remote.gs import RemoteGS
        return RemoteGS(storage_config)
    if storage_config.scheme == 's3':
        from plynx.utils.remote.s3 import RemoteS3
        return RemoteS3(storage_config)
    else:
        raise Exception('Unknown scheme `{}`'.format(storage_config.scheme))



import logging


def set_logging_level(verbose):
    levels = {
        0: logging.CRITICAL,
        1: logging.ERROR,
        2: logging.WARNING,
        3: logging.INFO,
        4: logging.DEBUG
    }
    root = logging.getLogger()
    root.setLevel(levels.get(verbose, logging.DEBUG))



from flask import g
from plynx.db.demo_user_manager import DemoUserManager
from plynx.web.common import app, requires_auth, make_fail_response, handle_errors
from plynx.utils.common import JSONEncoder
from plynx.constants import Collections


demo_user_manager = DemoUserManager()


@app.route('/plynx/api/v0/token', strict_slashes=False)
@requires_auth
@handle_errors
def get_auth_token():
    access_token = g.user.generate_access_token()
    refresh_token = g.user.generate_refresh_token()
    return JSONEncoder().encode({
        'access_token': access_token.decode('ascii'),
        'refresh_token': refresh_token.decode('ascii')
    })


@app.route('/plynx/api/v0/demo', methods=['POST'])
@handle_errors
def post_demo_user():
    user = demo_user_manager.create_demo_user()
    if not user:
        return make_fail_response('Failed to create a demo user')

    access_token = user.generate_access_token(expiration=1800)
    return JSONEncoder().encode({
        'access_token': access_token.decode('ascii'),
        'refresh_token': 'Not assigned',
        'username': user.username,
        'url': '/{}/{}'.format(Collections.TEMPLATES, DemoUserManager.demo_config.kind),
    })



from plynx.db.node import Node, Input, Parameter, Output
from plynx.constants import ParameterTypes


def get_test_node():
    node = Node()
    node.title = 'Command 1x1'
    node.description = 'Any command with 1 arg'
    node.kind = "dummy"

    node.inputs = []
    node.inputs.append(Input())
    node.inputs[-1].name = 'in'
    node.inputs[-1].file_type = 'file'
    node.inputs[-1].values = []

    node.outputs = []
    node.outputs.append(Output())
    node.outputs[-1].name = 'out'
    node.outputs[-1].file_type = 'file'

    node.parameters = []
    node.parameters.append(Parameter())
    node.parameters[-1].name = 'number'
    node.parameters[-1].parameter_type = ParameterTypes.INT
    node.parameters[-1].value = -1
    node.parameters[-1].widget = 'Number'

    node.parameters.append(Parameter())
    node.parameters[-1].name = 'cmd'
    node.parameters[-1].parameter_type = ParameterTypes.STR
    node.parameters[-1].value = 'cat ${input[in]} | grep ${param[text]} > ${output[out]}'
    node.parameters[-1].widget = 'Command line'

    return node


def compare_dictionaries(dict1, dict2):
    if dict1 is None or dict2 is None:
        return True

    if not isinstance(dict1, dict) or not isinstance(dict2, dict):
        return False

    shared_keys = set(dict2.keys()) & set(dict2.keys())

    if not (len(shared_keys) == len(dict1.keys()) and len(shared_keys) == len(dict2.keys())):
        return False

    dicts_are_equal = True
    for key in dict1.keys():
        if isinstance(dict1[key], dict):
            dicts_are_equal = dicts_are_equal and compare_dictionaries(dict1[key], dict2[key])
        else:
            dicts_are_equal = dicts_are_equal and (dict1[key] == dict2[key])

    return dicts_are_equal


def test_serialization():
    node1 = get_test_node()
    node1_dict = node1.to_dict()
    node2 = Node.from_dict(node1_dict)
    node2_dict = node2.to_dict()

    print(node1_dict)
    print("-")
    print(node2_dict)

    assert compare_dictionaries(node1_dict, node2_dict), "Serialized nodes are not equal"



import json
from flask import request, send_file, g
# !!! TODO figure out with STATIC
from plynx.web.common import app, requires_auth, make_fail_response, handle_errors
import plynx.base.resource
from plynx.plugins.resources.common import FILE_KIND
import plynx.utils.plugin_manager
from plynx.utils.common import JSONEncoder
from plynx.utils.file_handler import get_file_stream, upload_file_stream
from plynx.constants import ResourcePostStatus


RESOURCE_TYPES = list(plynx.utils.plugin_manager.get_resource_manager().kind_to_resource_class.keys())


@app.route('/plynx/api/v0/resource/<resource_id>', methods=['GET'])
@handle_errors
def get_resource(resource_id):
    preview = json.loads(request.args.get('preview', 'false'))
    file_type = request.args.get('file_type', None)
    if preview and not file_type:
        return make_fail_response('In preview mode `file_type` must be specified'), 400
    fp = get_file_stream(resource_id, preview=preview, file_type=file_type)
    if preview:
        preview_object = plynx.base.resource.PreviewObject(
            fp=fp,
            resource_id=resource_id,
        )
        return plynx.utils.plugin_manager.get_resource_manager().kind_to_resource_class[file_type].preview(preview_object)
    return send_file(
        fp,
        attachment_filename=resource_id)


@app.route('/plynx/api/v0/resource', methods=['POST'])
@handle_errors
@requires_auth
def post_resource():
    resource_id = upload_file_stream(request.files['data'])
    return JSONEncoder().encode({
        'status': ResourcePostStatus.SUCCESS,
        'resource_id': resource_id
    })


@app.route('/plynx/api/v0/upload_file', methods=['POST', 'PUT'])
@handle_errors
@requires_auth
def upload_file():
    assert len(request.files) == 1
    title = request.form.get('title', '{title}')
    description = request.form.get('description', '{description}')
    file_type = request.form.get('file_type', FILE_KIND)
    if file_type not in RESOURCE_TYPES:
        return make_fail_response('Unknown file type `{}`'.format(file_type)), 400

    resource_id = upload_file_stream(request.files['data'])
    raise NotImplementedError()
    file = None
    file.author = g.user._id
    file.title = title
    file.description = description
    file.outputs[0].resource_id = resource_id
    file.outputs[0].file_type = file_type
    file.save()

    return JSONEncoder().encode({
        'status': ResourcePostStatus.SUCCESS,
        'resource_id': resource_id,
        'node': file.to_dict(),
    })



from plynx.db.worker_state import get_worker_states
from plynx.utils.common import JSONEncoder
import plynx.utils.plugin_manager
from plynx.web.common import app, requires_auth, make_fail_response, handle_errors


PLUGINS_DICT = plynx.utils.plugin_manager.get_plugins_dict()


@app.route('/plynx/api/v0/worker_states', methods=['GET', 'POST'])
@handle_errors
@requires_auth
def worker_states():
    try:
        return JSONEncoder().encode({
            'status': 'success',
            'items': get_worker_states(),
            'plugins_dict': PLUGINS_DICT,
        })
    except Exception as e:
        app.logger.error(e)
        return make_fail_response('Internal error: "{}"'.format(str(e)))



from collections import namedtuple


class WorkerMessageType:
    """Basic Massage type from Worker"""
    UNKNOWN = 0
    HEARTBEAT = 1
    GET_JOB = 2
    JOB_FINISHED_SUCCESS = 3
    JOB_FINISHED_FAILED = 4


class RunStatus:
    """Worker running status"""
    IDLE = 0
    RUNNING = 1
    SUCCESS = 2
    FAILED = 3


class MasterMessageType:
    """Basic Massage type from Master"""
    UNKNOWN = 0
    AKNOWLEDGE = 1
    SET_JOB = 2
    KILL = 3


WorkerMessage = namedtuple('WorkerMessage', ['worker_id', 'message_type', 'run_status', 'body', 'graph_id'])

MasterMessage = namedtuple('MasterMessage', ['worker_id', 'message_type', 'job', 'graph_id'])



import logging
import traceback
from flask import Flask, request, g, Response
from flask_cors import CORS
from functools import wraps
from plynx.db.user import User
from plynx.utils.config import get_config
from plynx.utils.common import JSONEncoder
from plynx.utils.db_connector import check_connection

app = Flask(__name__)

DEFAULT_USERNAME = 'default'
DEFAULT_PASSWORD = ''

_config = None


def verify_password(username_or_token, password):
    if _config and _config.auth.secret_key and username_or_token == DEFAULT_USERNAME:
        return False
    user = User.verify_auth_token(username_or_token)
    if not user:
        # try to authenticate with username/password
        user = User.find_user_by_name(username_or_token)
        if not user or not user.verify_password(password):
            return False
    g.user = user
    return True


def authenticate():
    # 401 response
    return Response(
        'Could not verify your access level for that URL; You have to login with proper credentials',
        401,
        {'WWW-Authenticate': 'PlynxBasic realm="Login Required"'}
    )


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        auth = request.authorization
        if not auth:
            # if auth not provided, try default
            auth = {'username': DEFAULT_USERNAME, 'password': DEFAULT_PASSWORD}
        if not verify_password(auth['username'], auth['password']):
            return authenticate()
        return f(*args, **kwargs)
    return decorated


def register_user(username, password):
    """Register a new user.

    Args:
        username    (str):  Username
        password    (str):  Pasword

    Return:
        (str):  None if success, or error message if failed
    """
    if username is None or password is None:
        return 'Missing username or password'

    if User.find_user_by_name(username):
        return 'User with name `{}` already exists'.format(username)

    user = User()
    user.username = username
    user.hash_password(password)
    user.save()
    return None


def _init_default_user():
    if not User.find_user_by_name(DEFAULT_USERNAME):
        message = register_user(DEFAULT_USERNAME, DEFAULT_PASSWORD)
        if message:
            raise Exception(message)
        logging.info('Created default user `{}`'.format(DEFAULT_USERNAME))
    else:
        logging.info('Default user `{}` already exists'.format(DEFAULT_USERNAME))


def make_fail_response(message):
    return JSONEncoder().encode({
        'status': 'failed',
        'message': message
    })


def make_success_response(message='Completed', **extra_response):
    return JSONEncoder().encode(dict(
        {
            'status': 'success',
            'message': message,
        }, **extra_response))


def handle_errors(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            app.logger.error(traceback.format_exc())
            return make_fail_response('Internal error: "{}"'.format(repr(e))), 500
    return decorated


def run_api():
    global _config
    _config = get_config()

    check_connection()

    # Create default user if in single user mode
    if not _config.auth.secret_key:
        logging.info('Single user mode')
        _init_default_user()
    else:
        logging.info('Multiple user mode')

    CORS(app, resources={r"/*": {"origins": "*"}})
    app.run(
        host=_config.web.host,
        port=_config.web.port,
        debug=_config.web.debug,
        use_reloader=False,
    )



from passlib.apps import custom_app_context as pwd_context
from plynx.constants import Collections
from plynx.db.db_object import DBObject, DBObjectField
from plynx.utils.db_connector import get_db_connector
from plynx.utils.common import ObjectId
from plynx.utils.config import get_auth_config
from itsdangerous import (SignatureExpired, BadSignature,
                          TimedJSONWebSignatureSerializer as TimedSerializer,
                          JSONWebSignatureSerializer as Serializer)


class User(DBObject):
    """Basic User class with db interface."""

    FIELDS = {
        '_id': DBObjectField(
            type=ObjectId,
            default=ObjectId,
            is_list=False,
            ),
        'username': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'password_hash': DBObjectField(
            type=str,
            default='',
            is_list=False,
            ),
        'active': DBObjectField(
            type=bool,
            default=True,
            is_list=False,
            ),
    }

    DB_COLLECTION = Collections.USERS

    def hash_password(self, password):
        """Change password.

        Args:
            password    (str)   Real password string
        """
        self.password_hash = pwd_context.encrypt(password)

    def verify_password(self, password):
        """Verify password.

        Args:
            password    (str)   Real password string

        Return:
            (bool)    True if password matches else False
        """
        return pwd_context.verify(password, self.password_hash)

    def generate_access_token(self, expiration=600):
        """Generate access token.

        Args:
            expiration  (int)   Time to Live (TTL) in sec

        Return:
            (str)   Secured token
        """
        s = TimedSerializer(get_auth_config().secret_key, expires_in=expiration)
        return s.dumps({'username': self.username, 'type': 'access'})

    def generate_refresh_token(self):
        """Generate refresh token.

        Return:
            (str)   Secured token
        """
        s = Serializer(get_auth_config().secret_key)
        return s.dumps({'username': self.username, 'type': 'refresh'})

    def __str__(self):
        return 'User(_id="{}", username={})'.format(self._id, self.username)

    def __repr__(self):
        return 'User({})'.format(self.to_dict())

    def __getattr__(self, name):
        raise Exception("Can't get attribute '{}'".format(name))

    @staticmethod
    def find_user_by_name(username):
        """Find User.

        Args:
            username    (str)   Username

        Return:
            (User)   User object or None
        """
        user_dict = getattr(get_db_connector(), User.DB_COLLECTION).find_one({'username': username})
        if not user_dict:
            return None

        return User(user_dict)

    @staticmethod
    def find_users():
        return getattr(get_db_connector(), User.DB_COLLECTION).find({})

    @staticmethod
    def verify_auth_token(token):
        """Verify token.

        Args:
            token   (str)   Token

        Return:
            (User)   User object or None
        """
        s = TimedSerializer(get_auth_config().secret_key)
        try:
            data = s.loads(token)
            if data['type'] != 'access':
                raise Exception('Not access token')
        except (BadSignature, SignatureExpired):
            # access token is not valid or expired
            s = Serializer(get_auth_config().secret_key)
            try:
                data = s.loads(token)
                if data['type'] != 'refresh':
                    raise Exception('Not refresh token')
            except Exception:
                return None
        except Exception as e:
            print("Unexpected exception: {}".format(e))
            return None
        user = User.find_user_by_name(data['username'])
        if not user.active:
            return None
        return user



import logging
import pydoc
from plynx.utils.config import get_plugins


def _isinstance_namedtuple(x):
    return isinstance(x, tuple) and getattr(x, '_fields', None) is not None


def _as_dict(obj):
    if _isinstance_namedtuple(obj):
        return {
            key: _as_dict(getattr(obj, key)) for key in obj._fields
        }
    elif isinstance(obj, list):
        return [_as_dict(value) for value in obj]
    return obj


class _ResourceManager(object):
    def __init__(self, plugins):
        self.resources = plugins.resources
        self.kind_to_resource_class = {
            resource.kind: pydoc.locate(resource.cls) for resource in self.resources
        }
        self.kind_to_resource_dict = {
            resource.kind: resource._asdict() for resource in self.resources
        }


class _ExecutorManager(object):
    def __init__(self, plugins):
        self.kind_to_executor_class = {}
        self.kind_to_icon = {}
        self.kind_to_color = {}
        self.kind_to_title = {}
        for o_or_w in plugins.workflows + plugins.operations + plugins.dummy_operations:
            self.kind_to_executor_class[o_or_w.kind] = pydoc.locate(o_or_w.executor)
            if not self.kind_to_executor_class[o_or_w.kind]:
                raise Exception('Executor `{}` not found'.format(o_or_w.executor))
            self.kind_to_icon[o_or_w.kind] = o_or_w.icon
            self.kind_to_color[o_or_w.kind] = o_or_w.color
            self.kind_to_title[o_or_w.kind] = o_or_w.title

        self.kind_info = {}
        for kind, executor_class in self.kind_to_executor_class.items():
            logging.warning("Initializing executor `{}`".format(kind))
            self.kind_info[kind] = {
                'is_graph': executor_class.IS_GRAPH,
                'title': self.kind_to_title[kind],
                'icon': self.kind_to_icon[kind],
                'color': self.kind_to_color[kind],
            }


class _OperationManager(object):
    def __init__(self, plugins):
        self.operations = plugins.operations
        self.kind_to_operation_dict = {
            operation.kind: _as_dict(operation) for operation in self.operations
        }


class _HubManager(object):
    def __init__(self, plugins):
        self.hubs = plugins.hubs
        self.kind_to_hub_class = {
            hub.kind: pydoc.locate(hub.cls)(**hub.args) for hub in self.hubs
        }
        self.kind_to_hub_dict = {
            hub.kind: hub._asdict() for hub in self.hubs
        }


class _WorkflowManager(object):
    def __init__(self, plugins):
        self.workflows = plugins.workflows
        self.kind_to_workflow_dict = {
            workflow.kind: workflow._asdict() for workflow in self.workflows
        }


_plugins = get_plugins()

_resource_manager = None
_operation_manager = None
_hub_manager = None
_workflow_manager = None
_executor_manager = None

_plugins_dict = None


def get_resource_manager():
    global _resource_manager
    if not _resource_manager:
        _resource_manager = _ResourceManager(_plugins)
    return _resource_manager


def get_operation_manager():
    global _operation_manager
    if not _operation_manager:
        _operation_manager = _OperationManager(_plugins)
    return _operation_manager


def get_hub_manager():
    global _hub_manager
    if not _hub_manager:
        _hub_manager = _HubManager(_plugins)
    return _hub_manager


def get_workflow_manager():
    global _workflow_manager
    if not _workflow_manager:
        _workflow_manager = _WorkflowManager(_plugins)
    return _workflow_manager


def get_executor_manager():
    global _executor_manager
    if not _executor_manager:
        _executor_manager = _ExecutorManager(_plugins)
    return _executor_manager


def get_plugins_dict():
    global _plugins_dict
    if not _plugins_dict:
        _plugins_dict = {
            'resources_dict': get_resource_manager().kind_to_resource_dict,
            'operations_dict': get_operation_manager().kind_to_operation_dict,
            'hubs_dict': get_hub_manager().kind_to_hub_dict,
            'workflows_dict': get_workflow_manager().kind_to_workflow_dict,
            'executors_info': get_executor_manager().kind_info,
        }
    return _plugins_dict



import logging
import time
from collections import defaultdict
from plynx.constants import ParameterTypes
from plynx.db.node import Node, Parameter
from plynx.db.validation_error import ValidationError
from plynx.constants import JobReturnStatus, NodeRunningStatus, ValidationTargetType, ValidationCode, SpecialNodeId, Collections
from plynx.utils.common import to_object_id
import plynx.base.executor
import plynx.utils.executor
import plynx.db.node_collection_manager
import plynx.db.run_cancellation_manager


node_collection_manager = plynx.db.node_collection_manager.NodeCollectionManager(collection=Collections.RUNS)
run_cancellation_manager = plynx.db.run_cancellation_manager.RunCancellationManager()

_GRAPH_ITERATION_SLEEP = 1
_WAIT_STATUS_BEFORE_FAILED = {
    NodeRunningStatus.RUNNING,
    NodeRunningStatus.IN_QUEUE,
    NodeRunningStatus.FAILED_WAITING,
}
_ACTIVE_WAITING_TO_STOP = {
    NodeRunningStatus.FAILED_WAITING,
    NodeRunningStatus.CANCELED,
}


class DAG(plynx.base.executor.BaseExecutor):
    """ Main graph scheduler.

    Args:
        node_dict (dict)

    """
    IS_GRAPH = True

    def __init__(self, node_dict):
        super(DAG, self).__init__(node_dict)

        self.subnodes = None
        # TODO make a function to look for parameter
        for parameter in self.node.parameters:
            if parameter.name == '_nodes':
                self.subnodes = parameter.value.value

        assert self.subnodes is not None, 'Could not find subnodes'

        self.node_id_to_node = {
            node._id: node for node in self.subnodes
        }

        # number of dependencies to ids
        self.dependency_index_to_node_ids = defaultdict(lambda: set())
        self.node_id_to_dependents = defaultdict(lambda: set())
        self.node_id_to_dependency_index = defaultdict(lambda: 0)
        self.uncompleted_nodes_count = 0

        for node in self.subnodes:
            node_id = node._id
            if node_id == SpecialNodeId.INPUT:
                updated_resources_count = 0
                for output in node.outputs:
                    for input in self.node.inputs:
                        if input.name == output.name:
                            updated_resources_count += 1
                            output.values = input.values
                if updated_resources_count != len(self.node.inputs):
                    raise Exception('Used {} inputs for {} outputs'.format(updated_resources_count, len(self.node.inputs)))

            # ignore nodes in finished statuses
            if NodeRunningStatus.is_finished(node.node_running_status) and node_id != SpecialNodeId.OUTPUT:
                continue
            dependency_index = 0
            for node_input in node.inputs:
                for input_reference in node_input.input_references:
                    dep_node_id = to_object_id(input_reference.node_id)
                    self.node_id_to_dependents[dep_node_id].add(node_id)
                    if not NodeRunningStatus.is_finished(self.node_id_to_node[dep_node_id].node_running_status):
                        dependency_index += 1

            if not NodeRunningStatus.is_finished(node.node_running_status):
                self.uncompleted_nodes_count += 1
            self.dependency_index_to_node_ids[dependency_index].add(node_id)
            self.node_id_to_dependency_index[node_id] = dependency_index

        self.monitoring_node_ids = set()

    def finished(self):
        if self.node.node_running_status in _ACTIVE_WAITING_TO_STOP:
            # wait for the rest of the running jobs to finish
            # check running status of each of the nodes
            for node in self.subnodes:
                if node.node_running_status in _WAIT_STATUS_BEFORE_FAILED:
                    return False

            # set status to FAILED
            if self.node.node_running_status == NodeRunningStatus.FAILED_WAITING:
                self.node.node_running_status = NodeRunningStatus.FAILED
            return True
        return self.node.node_running_status in {NodeRunningStatus.SUCCESS, NodeRunningStatus.FAILED, NodeRunningStatus.CANCELED}

    def pop_jobs(self):
        """Get a set of nodes with satisfied dependencies"""
        res = []
        logging.info("Pop jobs")

        for running_node_dict in node_collection_manager.get_db_objects_by_ids(self.monitoring_node_ids):
            # check status
            if NodeRunningStatus.is_finished(running_node_dict['node_running_status']):
                node = Node.from_dict(running_node_dict)
                self.update_node(node)
                self.monitoring_node_ids.remove(node._id)

        if NodeRunningStatus.is_failed(self.node.node_running_status):
            logging.info("Job in DAG failed, pop_jobs returns []")
            return res

        cached_nodes = []
        for node_id in self.dependency_index_to_node_ids[0]:
            """Get the node and init its inputs, i.e. filling its resource_ids"""
            orig_node = self.node_id_to_node[node_id]
            for node_input in orig_node.inputs:
                for input_reference in node_input.input_references:
                    node_input.values.extend(
                        self.node_id_to_node[to_object_id(input_reference.node_id)].get_output_by_name(
                            input_reference.output_id
                        ).values
                    )
            orig_node.node_running_status = NodeRunningStatus.IN_QUEUE
            node = orig_node.copy()

            if DAG._cacheable(node) and False:  # !!! _cacheable is broken
                try:
                    cache = self.node_cache_manager.get(node, self.graph.author)
                    if cache:
                        node.node_running_status = NodeRunningStatus.RESTORED
                        node.outputs = cache.outputs
                        node.logs = cache.logs
                        node.cache_url = '{}/graphs/{}?nid={}'.format(
                            self.WEB_CONFIG.endpoint.rstrip('/'),
                            str(cache.graph_id),
                            str(cache.node_id),
                        )
                        cached_nodes.append(node)
                        continue
                except Exception as err:
                    logging.exception("Unable to update cache: `{}`".format(err))
            res.append(node)
        del self.dependency_index_to_node_ids[0]

        for node in cached_nodes:
            self.update_node(node)

        return res

    def update_node(self, node):
        dest_node = self.node_id_to_node[node._id]
        """
        if node.node_running_status == NodeRunningStatus.SUCCESS \
                and dest_node.node_running_status != node.node_running_status \
                and GraphScheduler._cacheable(node):
            GraphScheduler.node_cache_manager.post(node, self.graph_id, self.graph.author)
        """
        if dest_node.node_running_status == node.node_running_status:
            return

        self._set_node_status(node._id, node.node_running_status)
        # TODO smarter copy
        dest_node.parameters = node.parameters
        dest_node.logs = node.logs
        dest_node.outputs = node.outputs
        dest_node.cache_url = node.cache_url

    def _set_node_status(self, node_id, node_running_status):
        node = self.node_id_to_node[node_id]
        node.node_running_status = node_running_status

        logging.info("Node running status {} {}".format(node_running_status, node.title))

        if node_running_status == NodeRunningStatus.FAILED:
            # TODO optional cancel based on parameter
            self.kill()
            self.node.node_running_status = NodeRunningStatus.FAILED_WAITING

        if node_running_status in {NodeRunningStatus.SUCCESS, NodeRunningStatus.RESTORED}:
            for dependent_node_id in self.node_id_to_dependents[node_id]:
                dependent_node = self.node_id_to_node[dependent_node_id]
                prev_dependency_index = self.node_id_to_dependency_index[dependent_node_id]

                removed_dependencies = 0
                for node_input in dependent_node.inputs:
                    for input_reference in node_input.input_references:
                        if to_object_id(input_reference.node_id) == to_object_id(node_id):
                            removed_dependencies += 1
                dependency_index = prev_dependency_index - removed_dependencies

                self.dependency_index_to_node_ids[prev_dependency_index].remove(dependent_node_id)
                self.dependency_index_to_node_ids[dependency_index].add(dependent_node_id)
                self.node_id_to_dependency_index[dependent_node_id] = dependency_index
            self.uncompleted_nodes_count -= 1

        if self.uncompleted_nodes_count == 0 and not NodeRunningStatus.is_failed(self.node.node_running_status):
            self.node.node_running_status = NodeRunningStatus.SUCCESS

    @staticmethod
    def _cacheable(node):
        for parameter in node.parameters:
            if parameter.name == '_cacheable':
                return parameter.value
        return False

    @classmethod
    def get_default_node(cls, is_workflow):
        node = super().get_default_node(is_workflow)
        node.parameters.extend(
            [
                Parameter.from_dict({
                    'name': '_cacheable',
                    'parameter_type': ParameterTypes.BOOL,
                    'value': False,
                    'mutable_type': False,
                    'publicable': False,
                    'removable': False,
                }),
                Parameter.from_dict({
                    'name': '_timeout',
                    'parameter_type': ParameterTypes.INT,
                    'value': 600,
                    'mutable_type': False,
                    'publicable': True,
                    'removable': False
                }),
            ]
        )
        node.title = 'New DAG workflow'
        return node

    def _execute_node(self, node):
        if NodeRunningStatus.is_finished(node.node_running_status):     # NodeRunningStatus.SPECIAL
            return
        node.save(collection=Collections.RUNS)

        self.monitoring_node_ids.add(node._id)

    def run(self):
        while not self.finished():
            new_jobs = self.pop_jobs()
            if len(new_jobs) == 0:
                time.sleep(_GRAPH_ITERATION_SLEEP)
                continue

            for node in new_jobs:
                self._execute_node(node)

        is_succeeded = NodeRunningStatus.is_succeeded(self.node.node_running_status)
        if is_succeeded:
            for node in self.subnodes:
                if node._id != SpecialNodeId.OUTPUT:
                    continue
                updated_resources_count = 0
                for output in self.node.outputs:
                    for input in node.inputs:
                        if input.name == output.name:
                            output.values = input.values
                            updated_resources_count += 1
                            break
                if updated_resources_count != len(node.inputs):
                    raise Exception('Used {} inputs for {} outputs'.format(updated_resources_count, len(node.inputs)))
        return JobReturnStatus.SUCCESS if is_succeeded else JobReturnStatus.FAILED

    def kill(self):
        """Force to kill the process.

        The reason can be the fact it was working too long or parent exectuter canceled it.
        """
        self.node.node_running_status = NodeRunningStatus.CANCELED
        for node_id in list(self.monitoring_node_ids):
            run_cancellation_manager.cancel_run(node_id)

    def validate(self):
        validation_error = super().validate()
        if validation_error:
            return validation_error

        violations = []
        sub_nodes = self.node.get_parameter_by_name('_nodes').value.value

        if len(sub_nodes) == 0:
            violations.append(
                ValidationError(
                    target=ValidationTargetType.GRAPH,
                    object_id=str(self.node._id),
                    validation_code=ValidationCode.EMPTY_GRAPH
                ))

        for node in sub_nodes:
            node_violation = plynx.utils.executor.materialize_executor(node.to_dict()).validate()
            if node_violation:
                violations.append(node_violation)

        if len(violations) == 0:
            return None

        return ValidationError(
            target=ValidationTargetType.GRAPH,
            object_id=str(self.node._id),
            validation_code=ValidationCode.IN_DEPENDENTS,
            children=violations
        )



"""
The class defines `DBObject` and `DBObjectField`. This is an abstraction of all of the objects in database.
"""
import datetime
from collections import namedtuple
from plynx.utils.db_connector import get_db_connector
from plynx.utils.common import ObjectId

DBObjectField = namedtuple('DBObjectField', ['type', 'default', 'is_list'])

_registry = {}


def register_class(target_class):
    _registry[target_class.__name__] = target_class


def get_class(name):
    return _registry[name]


class DBObjectNotFound(Exception):
    """Internal Exception."""
    pass


class ClassNotSavable(Exception):
    """Internal Exception."""
    pass


class _DBObject(object):
    """DB Object.
    Abstraction of an object in the DB.

    Args:
        obj_dict    (dict, None):   Representation of the object. If None, an object with default fields will be created.
    """

    # Describe all the fields in the Object
    FIELDS = {}
    # Name of the collection in the database
    DB_COLLECTION = ''

    def __init__(self, obj_dict=None):
        self.__init_fields(obj_dict)
        self._dirty = True

    def is_dirty(self):
        return self._dirty

    def __setattr__(self, key, value):
        self.__dict__['_dirty'] = True
        self.__dict__[key] = value

    def __init_fields(self, obj_dict):
        obj_dict = obj_dict or {}
        for field_name, object_field in self.FIELDS.items():
            obj_value = obj_dict.get(field_name, None)
            if obj_value is not None:
                if object_field.is_list:
                    value = [
                        object_field.type(v) for v in obj_value
                    ]
                else:
                    value = object_field.type(obj_value)
            else:
                # Use default value
                value = object_field.default() if callable(object_field.default) else object_field.default
            setattr(self, field_name, value)

    @classmethod
    def load(cls, _id, collection=None):
        """Load object from db.

        Args:
            _id     (str, ObjectId):    ID of the object in DB
        """
        collection = collection or cls.DB_COLLECTION
        obj_dict = getattr(get_db_connector(), collection).find_one({'_id': ObjectId(_id)})
        if not obj_dict:
            raise DBObjectNotFound(
                'Object `{_id}` not found in `{collection}` collection'.format(
                    _id=_id,
                    collection=collection,
                )
            )
        return cls.from_dict(obj_dict)

    def save(self, force=False, collection=None):
        """Save Object in the database"""
        collection = collection or self.__class__.DB_COLLECTION
        if not collection:
            raise ClassNotSavable(
                "Class `{}` is not savable.".format(
                    self.__class__.__name__
                )
            )
        if not self.is_dirty() and not force:
            return True

        now = datetime.datetime.utcnow()

        obj_dict = self.to_dict()
        obj_dict["update_date"] = now

        getattr(get_db_connector(), collection).find_one_and_update(
            {'_id': obj_dict['_id']},
            {
                "$setOnInsert": {"insertion_date": now},
                "$set": obj_dict
            },
            upsert=True,
        )

        self._dirty = False
        return True

    @classmethod
    def from_dict(cls, obj_dict):
        """Create object from dict representation.

        Args:
            obj_dict    (dict):     Representation of the object,
                                    i.e. {'name': 'Obj Name', 'size': 12}
        """
        return cls(obj_dict)

    @staticmethod
    def __to_dict_single_element(value):
        if isinstance(value, DBObject):
            return value.to_dict()
        return value

    def to_dict(self):
        """Create a dictionary from an object."""
        res = {}
        for field_name, object_field in self.FIELDS.items():
            if object_field.is_list:
                value = [
                    self.__to_dict_single_element(element) for element in getattr(self, field_name)
                ]
            else:
                value = self.__to_dict_single_element(getattr(self, field_name))
            res[field_name] = value
        return res

    def copy(self):
        """Make a copy

        Return:
            A copy of the Object
        """
        return self.__class__(self.to_dict())

    def __str__(self):
        return '{cls_name}({value})'.format(
            cls_name=self.__class__.__name__,
            value=self.__dict__.get('_id', str(self.to_dict()))
        )

    def __repr__(self):
        return '{cls_name}({value})'.format(
            cls_name=self.__class__.__name__,
            value=str(self.to_dict()),
        )


class Meta(type):
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        register_class(cls)
        return cls


class DBObject(_DBObject, metaclass=Meta):
    pass



import sys
import csv
import logging
from datetime import datetime
import dateutil.parser
from collections import namedtuple
from plynx.db.node_cache import NodeCache
from plynx.db.node_cache_manager import NodeCacheManager
from plynx.utils.common import query_yes_no
from plynx.utils import file_handler


OutputListTuple = namedtuple('OutputListTuple', ['node_cache_id', 'insertion_date', 'resource_type', 'name', 'file_type', 'resource_id', 'protected'])

LIST_CACHE = 'list'
CLEAN_CACHE = 'clean'
MODES = [
    CLEAN_CACHE,
    LIST_CACHE,
]

node_cache_manager = NodeCacheManager()


def run_list_cache(start_datetime, end_datetime):
    file_writer = csv.writer(sys.stdout)
    file_writer.writerow(OutputListTuple._fields)
    for node_cache_dict in node_cache_manager.get_list(start_datetime, end_datetime):
        node_cache = NodeCache.from_dict(node_cache_dict)
        for resource_type in {'outputs', 'logs'}:
            for resource in getattr(node_cache, resource_type):
                if resource.resource_id:
                    file_writer.writerow(
                        OutputListTuple(
                            node_cache_id=node_cache._id,
                            insertion_date=node_cache_dict['insertion_date'].isoformat(),
                            resource_type=resource_type,
                            name=resource.name,
                            file_type=resource.file_type,
                            resource_id=resource.resource_id,
                            protected=node_cache.protected
                        )
                    )


def run_clean_cache(start_datetime, end_datetime, yes):
    query = node_cache_manager.get_list(start_datetime, end_datetime, non_protected_only=True)
    query_count = query.count()
    if query_count == 0:
        logging.critical('No cached results to remove')
        return 0
    if not yes and not query_yes_no('Are you sure you want to remove `{}` cached results?'.format(query_count), default='no'):
        print('Canceled')
        return 0

    logging.info('Start removing `{}` objects'.format(query_count))
    for node_cache_dict in query:
        node_cache = NodeCache.from_dict(node_cache_dict)
        try:
            for resource_type in {'outputs', 'logs'}:
                for resource in getattr(node_cache, resource_type):
                    if resource.resource_id:
                        file_handler.remove(resource.resource_id)
        # TODO use more cache states, such as `attempted to remove`
        finally:
            node_cache.removed = True
            node_cache.save()
            logging.info('Removed files from `{}`'.format(node_cache))

    node_cache_manager.clean_up()
    logging.info('Removed `{}` objects'.format(query_count))
    return 0


def run_cache(mode, start_datetime, end_datetime, yes):
    if mode not in MODES:
        raise ValueError('`mode` must be one of `{values}`. Value `{mode}` is given'.format(
            values=MODES,
            mode=mode,
        ))
    start_datetime = dateutil.parser.parse(start_datetime) if start_datetime else None
    end_datetime = dateutil.parser.parse(end_datetime) if end_datetime else datetime.now()
    if mode == LIST_CACHE:
        return run_list_cache(start_datetime, end_datetime)
    elif mode == CLEAN_CACHE:
        return run_clean_cache(start_datetime, end_datetime, yes)



#!/usr/bin/env python
import plynx
from setuptools import setup, find_packages

install_requires = [
    'pymongo>=3.9.0',
    'pyyaml>=5.1.2',
    'passlib>=1.7.1',
    'itsdangerous>=0.24',
    'Flask>=1.0.2',
    'Flask-Cors>=3.0.3',
    'Flask-HTTPAuth>=3.2.3',
    'gevent==1.3.7',
    'requests>=2.18.4',
    'future>=0.17.1',
]

# Extra dependencies for storage
gs = [
    "google-cloud-storage>=1.13.0",
]
s3 = [
    "boto3>=1.9.62",
]
ssh = [
    "paramiko>=2.4.2",
]
all_remotes = gs + s3 + ssh

setup(
    name='plynx',
    version=plynx.__version__,
    description='ML platform',
    long_description='Interactive, Scalable, Shareable and Reproducible ML experiments',
    url='https://plynx.com',
    author='Ivan Khomyakov',
    author_email='ivan@plynx.com',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Environment :: Console',
        'Environment :: Web Environment',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',

        'License :: OSI Approved :: Apache Software License',

        'Operating System :: POSIX',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',

        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
    ],
    keywords='data science, machine learning, pipeline, workflow, experiments',
    packages=find_packages(exclude=['scripts', 'docker']),
    install_requires=install_requires,
    extras_require={
        'all': all_remotes,
        'gs': gs,
        's3': s3,
        'ssh': ssh,
    },
    package_data={},
    entry_points={
        'console_scripts': [
            'plynx=plynx.bin:main',
        ],
    },
    project_urls={
        'Demo': 'https://plynx.com',
        'Source': 'https://github.com/plynx-team/plynx',
    },
    # plynx.graph.base_nodes.collection uses reference to __file__
    zip_safe=False,
)



import os
import json
import uuid
from plynx.constants import NodeResources
from plynx.base import resource
from plynx.utils.config import get_cloud_service_config


CLOUD_SERVICE_CONFIG = get_cloud_service_config()


class CloudStorage(resource.BaseResource):
    @staticmethod
    def prepare_input(filename, preview):
        if preview:
            uniq_id = str(uuid.uuid1())
            cloud_filename = os.path.join(
                '{prefix}/{workdir}'.format(
                    prefix=CLOUD_SERVICE_CONFIG.prefix,
                    workdir=uniq_id,
                )
            )
        else:
            with open(filename) as f:
                cloud_filename = json.load(f)['path']
        return {
            NodeResources.INPUT: filename,
            NodeResources.CLOUD_INPUT: cloud_filename,
        }

    def prepare_output(filename, preview):
        uniq_id = str(uuid.uuid1())
        cloud_filename = os.path.join(
            '{prefix}/{workdir}'.format(
                prefix=CLOUD_SERVICE_CONFIG.prefix,
                workdir=uniq_id,
            )
        )
        if not preview:
            with open(filename, 'w') as f:
                json.dump({"path": cloud_filename}, f)

        return {
                NodeResources.OUTPUT: filename,
                NodeResources.CLOUD_OUTPUT: cloud_filename,
            }

    @classmethod
    def preview(cls, preview_object):
        path = json.load(preview_object.fp)['path']
        return '<a href={}>{}</a>'.format(
            ''.join([CLOUD_SERVICE_CONFIG.url_prefix, path.split('//')[1], CLOUD_SERVICE_CONFIG.url_postfix]),
            path
        )
